from json import tool
from PySide2.QtWidgets import QWidget
from SciDataTool.Functions import parser
from PySide2.QtCore import Qt
from SciDataTool.GUI.DDataPlotter.Ui_DDataPlotter import Ui_DDataPlotter
from matplotlib.backends.backend_qt5agg import (
    FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar,
)
from SciDataTool.Functions.Plot.init_fig import init_fig
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.collections import PathCollection, QuadMesh
from matplotlib.text import Annotation
from numpy import array
from SciDataTool.Functions.is_axes_in_order import is_axes_in_order
from SciDataTool.Functions.Plot import TEXT_BOX


SYMBOL_DICT = {
    "time": "t",
    "angle": "\\alpha",
    "axial direction": "z",
    "frequency": "f",
    "wavenumber": "r",
    "rotation speed": "N0",
    "stator current along d-axis": "I_d",
    "stator current along q-axis": "I_q",
    "speed": "N0",
    "torque": "T",
    "a-weighted sound power level": "ASWL",
    "velocity level": "V",
    "reference torque": "T_{ref}",
}

PARAM_3D = [
    "is_2D_view",
    "is_contour",
    "is_same_size",
    "N_stem",
    "colormap",
    "annotation_delim",
    "marker_color",
    "z_range",
]

PARAM_2D = [
    "color_list",
    "data_list",
    "legend_list",
    "fund_harm_dict",
    "is_show_legend",
]


def latex(string):
    """format a string for latex"""
    if "_" in string or "^" in string or "\\" in string:
        string = r"$" + string + "$"
    return string


class DDataPlotter(Ui_DDataPlotter, QWidget):
    """Main window of the SciDataTool UI"""

    def __init__(
        self,
        data,
        axes_request_list=list(),
        component=None,
        unit=None,
        z_min=None,
        z_max=None,
        is_auto_refresh=False,
        frozen_type=0,
        plot_arg_dict=dict(),
        save_path="",
        logger=None,
        path_to_image=None,
    ):
        """Initialize the UI according to the input given by the user

        Parameters
        ----------
        self : DDataPlotter
            a DDataPlotter object
        data : DataND or VectorField object
            A DataND/VectorField object to plot
        axes_request_list: list
            list of RequestedAxis which are the info given for the autoplot (for the axes and DataSelection)
        component : str
            Name of the component to plot (For VectorField only)
        unit : str
            unit in which to plot the field
        z_min : float
            Minimum value for Z axis (or Y if only one axe)
        z_max : float
            Minimum value for Z axis (or Y if only one axe)
        is_auto_refresh : bool
            True to refresh at each widget changed (else wait call to button)
        plot_arg_dict : dict
            Dictionnary with arguments that must be given to the plot
        frozen_type : int
            0 to let the user modify the axis of the plot, 1 to let him switch them, 2 to not let him change them,
            3 to freeze both axes and operations, 4 to freeze fft, 5 to only allow switch
        save_path : str
            path to the folder where the animations are saved
        logger : logger
            logger used to print path to animation (if None using print instead)
        path_to_image : str
            path to the folder where the image for the animation button is saved
        """

        # Build the interface according to the .ui file
        QWidget.__init__(self)
        self.setupUi(self)

        self.auto_refresh = is_auto_refresh
        if is_auto_refresh:
            self.is_auto_refresh.setCheckState(Qt.Checked)
        else:
            self.is_auto_refresh.setCheckState(Qt.Unchecked)

        self.plot_arg_dict = plot_arg_dict
        self.data = data
        self.data_orig = data

        # Initializing the figure inside the UI
        (self.fig, self.ax, _, _) = init_fig()
        # self.set_figure(self.fig)

        if (
            "type_plot" in self.plot_arg_dict
            and self.plot_arg_dict["type_plot"] == "quiver"
        ):
            is_quiver = True
        else:
            is_quiver = False

        # Initializing the WPlotManager
        self.w_plot_manager.set_info(
            data=data,
            axes_request_list=axes_request_list,
            component=component,
            unit=unit,
            z_min=z_min,
            z_max=z_max,
            frozen_type=frozen_type,
            is_quiver=is_quiver,
            plot_arg_dict=plot_arg_dict,
            save_path=save_path,
            logger=logger,
            path_to_image=path_to_image,
            main_widget=self,
        )

        # Building the interaction with the UI itself
        self.b_refresh.clicked.connect(self.update_plot)
        self.w_plot_manager.updatePlot.connect(self.auto_update)
        self.w_plot_manager.updatePlotForced.connect(self.update_plot)
        self.update_plot()

        # Adding an argument for testing autorefresh
        self.is_plot_updated = False
        self.is_auto_refresh.toggled.connect(self.set_auto_refresh)

    def auto_update(self):
        """Method that checks if the autorefresh is enabled. If true, then it updates the plot.
        Parameters
        ----------
        self : WPlotManager
            a WPlotManager object

        """

        self.b_refresh.setDisabled(False)

        if self.auto_refresh == True:
            self.update_plot()
            self.is_plot_updated = True
        else:
            self.is_plot_updated = False

    def set_auto_refresh(self):
        """Method that update the refresh policy according to the checkbox inside the UI

        Parameters
        ----------
        self : WPlotManager
            a WPlotManager object

        """
        self.auto_refresh = self.is_auto_refresh.isChecked()

        if self.is_auto_refresh.isChecked():
            # When auto-refresh is enabled, the refresh button must be disabled
            self.b_refresh.setDisabled(True)
            # Refresh
            self.update_plot()
        else:
            # When auto-refresh is disabled, the refresh button must be enabled
            self.b_refresh.setDisabled(False)

    def set_figure(self, fig, text_box=None):
        """Method that set up the figure inside the GUI

        Parameters
        ----------
        self : DDataPlotter
            a DDataPlotter object
        fig : Figure
            A Figure object to put inside the UI

        """
        if text_box is None:
            text_box = TEXT_BOX
        # Set plot layout
        self.canvas = FigureCanvas(fig)
        if self.toolbar is None:
            toolbar = NavigationToolbar(self.canvas, self)
        else:
            toolbar = NavigationToolbar(self.canvas, self)
            action_names = [action.text() for action in self.toolbar.actions()]
            for action in toolbar.actions():
                if action.text() in action_names:
                    action.setIcon(
                        self.toolbar.actions()[action_names.index(action.text())].icon()
                    )
                else:
                    toolbar.removeAction(action)

        self.plot_layout.addWidget(toolbar)
        self.plot_layout.addWidget(self.canvas)

        self.text = None
        self.circle = None
        self.line = None

        # Set labels for cursor and coordinates
        ######################################
        def format_coord(x, y, z=None, sep=", ", ind=None):
            # Use hidden annotations
            annotations = [
                child
                for child in self.ax.get_children()
                if isinstance(child, Annotation)
            ]
            is_annot = True
            X_str = None
            if (
                ind is not None
                and annotations != []
                and ind[0] in range(len(annotations))
                and not annotations[ind[0]]._visible
            ):
                if annotations[ind[0]]._x == x:
                    X_str = annotations[ind[0]]._text
                    is_annot = False
            if is_annot:
                # Use ticklabels
                try:
                    x_float = float(self.ax.get_xticklabels()[-1]._text)
                    X_str = format(x, ".4g")
                except:
                    for ticklabel in self.ax.get_xticklabels():
                        if ticklabel._x == x:
                            X_str = ticklabel._text
                            break
            if "mathdefault" in self.ax.get_yticklabels()[-1]._text:
                Y_str = format(y, ".4g")
            else:
                try:
                    y_float = float(self.ax.get_yticklabels()[-1]._text)
                    Y_str = format(y, ".4g")
                except:
                    Y_str = None
                    for ticklabel in self.ax.get_yticklabels():
                        if ticklabel._y == y:
                            Y_str = ticklabel._text
                            break
            if X_str is None or Y_str is None:
                return ""

            # Recovering the input of the user
            [
                self.data,
                axes_selected,
                _,
                output_range,
            ] = self.w_plot_manager.get_plot_info()

            # Checking if the axes are following the order inside the data object
            axes_selected_parsed = parser.read_input_strings(
                axes_selected, axis_data=None
            )

            if axes_selected_parsed[0].name.lower() in SYMBOL_DICT:
                xlabel = latex(SYMBOL_DICT[axes_selected_parsed[0].name.lower()])
            else:
                xlabel = latex(axes_selected_parsed[0].name)
            xunit = "[" + axes_selected_parsed[0].unit + "]"

            if len(axes_selected) == 2:
                if axes_selected_parsed[1].name.lower() in SYMBOL_DICT:
                    ylabel = latex(SYMBOL_DICT[axes_selected_parsed[1].name.lower()])
                else:
                    ylabel = latex(axes_selected_parsed[1].name)
                yunit = "[" + axes_selected_parsed[1].unit + "]"

            else:
                if self.data.name.lower() in SYMBOL_DICT:
                    ylabel = latex(SYMBOL_DICT[self.data.name.lower()])
                else:
                    ylabel = latex(self.data.symbol)

                yunit = "[" + latex(output_range["unit"]) + "]"

                if ylabel == "W" and "dBA" in yunit:
                    ylabel = "ASWL"
                elif ylabel == "W" and "dB" in yunit:
                    ylabel = "SWL"

            # Keep only one decimal in dB case
            if "dB" in yunit:
                Y_str = format(float(Y_str), ".1f")

            label = (
                xlabel
                + " = "
                + X_str
                + " "
                + xunit
                + sep
                + ylabel
                + " = "
                + Y_str
                + " "
                + yunit
            )

            if z is not None:
                if hasattr(self, "data") and self.data is not None:
                    zlabel = latex(self.data.symbol)
                elif self.data._name.lower() in SYMBOL_DICT:
                    zlabel = latex(SYMBOL_DICT[self.data._name.lower()])
                else:
                    zlabel = latex(self.data._name)

                zunit = "[" + latex(output_range["unit"]) + "]"

                if zlabel == "W" and "dBA" in zunit:
                    zlabel = "ASWL"
                elif zlabel == "W" and "dB" in zunit:
                    zlabel = "SWL"

                # Keep only one decimal in dB case
                if "dB" in yunit:
                    Z_str = format(z, ".1f")
                else:
                    Z_str = format(z, ".4g")

                label += sep + zlabel + " = " + Z_str + " " + zunit

            # Add hidden annotations
            if annotations != [] and not annotations[0]._visible and is_annot:
                for annotation in annotations:
                    if annotation._x == x and annotation._y == y:
                        label += sep + annotation._text

            # Remove latex marks for top right corner
            if sep == ", ":
                label = label.replace("$", "").replace("{", "").replace("}", "")

            return label.replace("SI", "").replace(" []", "")  # Remove dimless units

        # Set cursor
        ######################################
        def set_cursor(event):
            plot_obj = event.artist
            Z = None
            legend = None
            annot = None
            if hasattr(event, "ind"):
                ind = event.ind
            else:
                ind = None
            if isinstance(plot_obj, Line2D):
                xdata = plot_obj.get_xdata()
                ydata = plot_obj.get_ydata()
                X = xdata[ind][0]  # X position of the click
                Y = ydata[ind][0]  # Y position of the click
                if self.ax.get_legend_handles_labels()[1] != []:
                    legend = self.ax.get_legend_handles_labels()[1][
                        self.ax.lines.index(plot_obj)
                    ]
                    annotations = [
                        child
                        for child in self.ax.get_children()
                        if isinstance(child, Annotation)
                    ]
                    if (
                        annotations != []
                        and self.ax.lines.index(plot_obj) in range(len(annotations))
                        and not annotations[self.ax.lines.index(plot_obj)]._visible
                        and "Overall" not in legend
                    ):
                        annot = annotations[self.ax.lines.index(plot_obj)]._text
            elif isinstance(plot_obj, PathCollection):
                X = plot_obj.get_offsets().data[ind][0][0]
                Y = plot_obj.get_offsets().data[ind][0][1]
                if plot_obj.get_array() is not None:
                    Z = plot_obj.get_array().data[ind][0]
            elif isinstance(plot_obj, Rectangle):
                X = plot_obj.get_x() + plot_obj.get_width() / 2
                Y = plot_obj.get_height()
            elif isinstance(plot_obj, QuadMesh):
                if plot_obj._coordinates.shape[0] > 2:
                    # pcolormesh case
                    l = ind[0] // plot_obj._coordinates.shape[1]
                    X = plot_obj._coordinates[
                        l, ind[0] - l * plot_obj._coordinates.shape[1] + 1, 0
                    ]
                    Y = plot_obj._coordinates[
                        l, ind[0] - l * plot_obj._coordinates.shape[1] + 1, 1
                    ]
                else:
                    X = (
                        plot_obj._coordinates[0, ind[0] + 1, 0]
                        + plot_obj._coordinates[1, ind[0] + 1, 0]
                    ) / 2
                    Y = (
                        plot_obj._coordinates[0, ind[0] + 1, 1]
                        + plot_obj._coordinates[1, ind[0] + 1, 1]
                    ) / 2
                Z = plot_obj.get_array().data[ind[0]]

            # Offset for the label
            x_min, x_max = self.ax.get_xlim()
            dx = (x_max - x_min) / 50
            if X is not None and Y is not None:
                if isinstance(X, str):
                    X = ind
                label = format_coord(X, Y, Z, sep="\n", ind=ind)
                if legend is not None:
                    label = legend + "\n" + label
                if annot is not None:
                    label += "\n" + annot
                if label != "":
                    if self.text is None:
                        # Create label in box and black cross
                        self.text = self.ax.text(
                            X + dx,
                            Y,
                            label,
                            ha="left",
                            va="center",
                            bbox=text_box,
                        )
                        # Draw line
                        self.line = self.ax.plot(
                            [X, X + dx],
                            [Y, Y],
                            color="k",
                            linestyle="-",
                            linewidth=0.5,
                        )
                        # Draw circle
                        self.circle = self.ax.plot(
                            X,
                            Y,
                            ".",
                            markerfacecolor="w",
                            markeredgecolor="k",
                            markeredgewidth=0.5,
                            markersize=12,
                        )
                    else:
                        # Update label, line and circle
                        self.text._x = X + dx
                        self.text._y = Y
                        self.text._text = label
                        self.circle[0].set_xdata(array(X))
                        self.circle[0].set_ydata(array(Y))
                        self.line[0].set_xdata(array([X, X + dx]))
                        self.line[0].set_ydata(array([Y, Y]))

                    self.ax.texts[-1].set_visible(True)
                    self.ax.lines[-1].set_visible(True)
                    self.ax.lines[-2].set_visible(True)
                    self.ax.zorder = 10
                    self.canvas.draw()

        def delete_cursor(event):
            if event.button.name == "RIGHT":
                if self.text is not None:
                    self.ax.texts[-1].set_visible(False)
                if self.line is not None:
                    self.ax.lines[-1].set_visible(False)
                if self.circle is not None:
                    self.ax.lines[-2].set_visible(False)
            self.canvas.draw()

        self.ax.format_coord = format_coord
        self.canvas.mpl_connect("pick_event", set_cursor)
        self.canvas.mpl_connect("button_press_event", delete_cursor)

    def update_plot(self):
        """Method that update the plot according to the info selected in the UI
        Parameters
        ----------
        self : DDataPlotter
            a DDataPlotter object

        """
        # Disabling refresh button after clicking on it (similar to * for an unsaved file)
        self.b_refresh.setDisabled(True)

        # Clear plots
        for i in reversed(range(self.plot_layout.count())):
            if self.plot_layout.itemAt(i).widget() is not None:
                widgetToRemove = self.plot_layout.itemAt(i).widget()
                widgetToRemove.deleteLater()

        if self.fig.get_axes():
            if len(self.fig.canvas.toolbar.actions()) == 11:
                self.toolbar = None
            else:
                self.toolbar = self.fig.canvas.toolbar
            self.fig.clear()

        (self.fig, self.ax, _, _) = init_fig()
        self.set_figure(self.fig)

        # Recovering the input of the user
        [
            self.data,
            axes_selected,
            data_selection,
            output_range,
        ] = self.w_plot_manager.get_plot_info()

        # Checking if the axes are following the order inside the data object
        not_in_order = False

        if len(axes_selected) == 2:
            not_in_order, axes_selected = is_axes_in_order(axes_selected, self.data)

        if not None in data_selection:
            if len(axes_selected) == 1:
                plot_arg_dict_2D = self.plot_arg_dict.copy()
                for param in PARAM_3D:
                    if param in plot_arg_dict_2D:
                        del plot_arg_dict_2D[param]
                if (
                    "type_plot" in plot_arg_dict_2D
                    and plot_arg_dict_2D["type_plot"] == "quiver"
                ):
                    if (
                        self.w_plot_manager.w_vect_selector.get_component_selected()
                        != "all"
                    ):
                        component_list = [
                            self.w_plot_manager.w_vect_selector.get_component_selected()
                        ]
                    else:
                        component_list = None
                    angle_str = "angle"
                    if "is_smallestperiod" in plot_arg_dict_2D:
                        if plot_arg_dict_2D["is_smallestperiod"]:
                            angle_str = "angle[smallestperiod]"
                        del plot_arg_dict_2D["is_smallestperiod"]
                    if "fig" in plot_arg_dict_2D:
                        del plot_arg_dict_2D["fig"]
                    self.data_orig.plot_2D_Data(
                        *[*[angle_str], *data_selection],
                        **plot_arg_dict_2D,
                        unit=output_range["unit"],
                        fig=self.fig,
                        ax=self.ax,
                        y_min=output_range["min"],
                        y_max=output_range["max"],
                        component_list=component_list,
                    )
                else:
                    if "axis_data" in plot_arg_dict_2D:
                        for axis_name in plot_arg_dict_2D["axis_data"]:
                            for i, axis_selected in enumerate(axes_selected):
                                if axis_name in axis_selected:
                                    if "axis_data" not in axis_selected:
                                        axes_selected[i] = (
                                            axis_selected.split("{")[0]
                                            + "=axis_data"
                                            + "{"
                                            + axis_selected.split("{")[1]
                                        )
                    if "title" in plot_arg_dict_2D and "[" in data_selection[-1]:
                        axis_cont = data_selection[-1].split("[")[0]
                        index = int(data_selection[-1].split("[")[1].split("]")[0])
                        plot_arg_dict_2D["title"] += (
                            " for "
                            + axis_cont.rstrip("s")
                            + " "
                            + self.data.get_axes(axis_cont)[0].values[index]
                        )
                    self.data.plot_2D_Data(
                        *[*axes_selected, *data_selection],
                        **plot_arg_dict_2D,
                        unit=output_range["unit"],
                        fig=self.fig,
                        ax=self.ax,
                        y_min=output_range["min"],
                        y_max=output_range["max"],
                    )

            elif len(axes_selected) == 2:
                plot_arg_dict_3D = self.plot_arg_dict.copy()
                for param in PARAM_2D:
                    if param in plot_arg_dict_3D:
                        del plot_arg_dict_3D[param]
                if "is_2D_view" not in self.plot_arg_dict:
                    self.plot_arg_dict["is_2D_view"] = True
                if "axis_data" in plot_arg_dict_3D:
                    for axis_name in plot_arg_dict_3D["axis_data"]:
                        for i, axis_selected in enumerate(axes_selected):
                            if axis_name in axis_selected:
                                if "axis_data" not in axis_selected:
                                    axes_selected[i] = (
                                        axis_selected.split("{")[0]
                                        + "=axis_data"
                                        + "{"
                                        + axis_selected.split("{")[1]
                                    )
                if output_range["min"] is not None:
                    plot_arg_dict_3D["z_min"] = output_range["min"]
                if output_range["max"] is not None:
                    plot_arg_dict_3D["z_max"] = output_range["max"]
                if "is_switch_axes" in plot_arg_dict_3D:
                    del plot_arg_dict_3D["is_switch_axes"]
                self.data.plot_3D_Data(
                    *[*axes_selected, *data_selection],
                    **plot_arg_dict_3D,
                    unit=output_range["unit"],
                    fig=self.fig,
                    ax=self.ax,
                    is_switch_axes=not_in_order,
                )

        else:
            print("Operation not implemented yet, plot could not be updated")

        self.w_plot_manager.w_range.set_min_max()

    def set_info(
        self,
        data,
        unit=None,
        axes_request_list=None,
        plot_arg_dict=dict(),
        is_keep_config=False,
        frozen_type=0,
        save_path="",
        logger=None,
        path_to_image=None,
    ):
        """Method to set the DDataPlotter with information given
        self : DDataPlotter
            a DDataPlotter object
        data : DataND or VectorField object
            A DataND/VectorField object to plot
        axes_request_list: list
            list of RequestedAxis which are the info given for the autoplot (for the axes and DataSelection)
        plot_arg_dict : dict
            Dictionnary with arguments that must be given to the plot
        frozen_type : int
            0 to let the user modify the axis of the plot, 1 to let him switch them, 2 to not let him change them, 3 to freeze both axes and operations, 4 to freeze fft
        save_path : str
            path to the folder where the animations are saved
        logger : logger
            logger used to print path to animation (if None using print instead)
        path_to_image : str
            path to the folder where the image for the animation button is saved
        """

        self.plot_arg_dict = plot_arg_dict
        if (
            "type_plot" in self.plot_arg_dict
            and self.plot_arg_dict["type_plot"] == "quiver"
        ):
            is_quiver = True
        else:
            is_quiver = False

        self.w_plot_manager.set_info(
            data=data,
            unit=unit,
            axes_request_list=axes_request_list,
            is_keep_config=is_keep_config,
            frozen_type=frozen_type,
            plot_arg_dict=plot_arg_dict,
            is_quiver=is_quiver,
            save_path=save_path,
            logger=logger,
            path_to_image=path_to_image,
        )

    def showEvent(self, ev):
        super(DDataPlotter, self).showEvent(ev)
        self.w_scroll.setFixedWidth(
            498 + self.w_scroll.verticalScrollBar().sizeHint().width()
        )
        self.w_scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)



from PySide2.QtWidgets import QLabel


class ButtonLabel(QLabel):
    """A QLabel Widget for buttons with hovering stylesheet"""

    def __init__(self, *args, **kwargs):
        """Same constructor as QLabel"""

        # Call the QLabel constructor
        super(ButtonLabel, self).__init__(*args, **kwargs)
        self.setMouseTracking(True)
        self.setStyleSheet(
            ".ButtonLabel{border: 1px solid transparent; border-radius: 6px;}"
        )

    def enterEvent(self, event):
        """To add border when hovering label"""
        self.setStyleSheet(
            ".ButtonLabel{border: 1px solid #0069A1; border-radius: 6px;}"
        )

    def leaveEvent(self, event):
        """To remove border when leaving label"""
        self.setStyleSheet(
            ".ButtonLabel{border: 1px solid transparent; border-radius: 6px;}"
        )



import numpy as np


def filter_spectral_leakage(self, freqs_th):
    """Filter spectral leakage from a SciDataTool VectorField object

    Parameters
    ----------
    self : VectorField
        a VectorField object to be filtered
    freqs_th : ndarray
        theoretical frequencies

    Returns
    -------
    vf_filt : VectorField
        Filtered vectorfield

    """

    # Init filtered VectorField
    vf_filt = type(self)(name=self.name, symbol=self.symbol, components=dict())

    axes_list = list()
    arg_list = list()
    Wmatf = np.array([])
    If = np.array([])
    for comp, data in self.components.items():
        # Filter components by components
        (
            data_filt,
            axes_list,
            arg_list,
            Wmatf,
            If,
            freqs_th,
        ) = data.filter_spectral_leakage(
            freqs_th, axes_list, arg_list, Wmatf, If, is_return_calc_data=True
        )
        # Store filtered data in VectorField
        vf_filt.components[comp] = data_filt

    return vf_filt



from numpy import array
from SciDataTool.Functions.Plot import fft_dict, ifft_dict


def fix_axes_order(axes_list, args):
    """Put args in same order as axes_list"""
    indices = []
    for j, arg in enumerate(args):
        for i, axis in enumerate(axes_list):
            if (
                axis in arg
                or axis in fft_dict
                and fft_dict[axis] in arg
                or axis in ifft_dict
                and ifft_dict[axis] in arg
            ):
                indices.append(i)
                break
    if len(indices) < len(args):
        # axis was not in data -> keep same order
        return args
    else:
        sort_indices = array(indices).argsort()
        return [args[i] for i in sort_indices]



import numpy as np
import numpy.linalg as np_lin

from SciDataTool.Functions.set_routines import unique_tol


def filter_spectral_leakage(
    spectrum,
    freqs,
    freqs_th,
    Nt,
    dt,
    is_freq_pos=True,
    Wmatf=np.array([]),
    If=np.array([]),
    is_plot=False,
):
    """Filter spectral leakage from the input spectrum and theoretical frequencies accordingly to the method developed in
    Rainer et al., "Weak Coupling Between Electromagnetic and Structural Models for Electrical Machines",
    IEEE TRANSACTIONS ON MAGNETICS, VOL. 46, NO. 8, AUGUST 2010.

    Parameters
    ----------
    spectrum : ndarray
        Frequency spectrum to filter (1D, 2D or 3D). In case of 2D/3D, angle axis can be wavenumber axis
    freqs_th : ndarray
        theoretical frequencies
    Nt : int
        Number of time steps
    dt: float
        Time step value
    is_freq_pos: bool
        True to return spectrum truncated to positive frequencies
    Wmatf : ndarray
        Filtering matrix
    If: ndarray
        Index of FFT harmonics components used to calculate Wmatf

    Returns
    -------
    spectrum_filt : ndarray
        Filtered spectrum
    freqs_th : ndarray
        theoretical frequencies
    Wmatf : ndarray
        Filtering matrix
    """

    # Symmetrize theoretical frequencies to negative values if necessary
    Nhalf = (
        int(freqs_th.size / 2)
        if freqs_th.size % 2 == 0
        else int((freqs_th.size + 1) / 2)
    )
    if freqs_th.size == 1 or not np.all(freqs_th[:Nhalf] == -np.flip(freqs_th[Nhalf:])):
        freqs_th = unique_tol(np.concatenate((freqs_th, -freqs_th), axis=0))

    if Wmatf.size == 0 or If.size == 0:

        # Check frequency resolution
        df_fft = np.min(np.abs(np.diff(freqs)))
        df_th = np.min(np.abs(np.diff(freqs_th)))
        if df_th < df_fft:
            print(
                "FFT frequency resolution lower than theoretical frequency resolution, spectral leakage filtering maybe inaccurate"
            )

        # Keep only theoretical frequencies in the calculated range
        I1 = np.abs(freqs_th) <= np.max(np.abs(freqs))
        freqs_th = freqs_th[I1]

        # Find closest index of each frequency in the grid
        If = np.argmin(np.abs(freqs[:, None] - freqs_th[None, :]), axis=0)

        # Filtering will be performed, match theoretical frequencies with closest
        # frequency in FFT frequency vector
        If0 = np.unique(If)
        if If0.size != If.size:
            If_new = np.zeros(freqs_th.size, dtype=int)
            for ii, jj in enumerate(If):
                kk = jj
                if kk + 1 < freqs.size:
                    # Going forward
                    while kk in If_new:
                        kk += 1
                else:
                    # Going backwards
                    while kk in If_new:
                        kk -= 1
                If_new[ii] = kk
            if np.unique(If_new).size < If.size:
                print(
                    "Wrong match between FFT and theoretical frequencies, spectral leakage filtering maybe inaccurate"
                )
            If = np.sort(If_new)

        # Calculate filtering matrix: spectrum of door window
        xfreqs2, xfreqs1 = np.meshgrid(freqs_th, freqs[If])
        Wmatf = my_doorwin(xfreqs1 - xfreqs2, Nt, dt, tol0=1e-4)

    # Check that spectrum is either 1D or 3D
    if spectrum.ndim > 2:
        # Reshape into 2D matrix
        Nangle = spectrum.shape[1]
        Nslice = spectrum.shape[2]
        spectrum = np.reshape(spectrum, (spectrum.shape[0], Nangle * Nslice))
        is_3D = True
    else:
        # Add 3rd dimension and squeeze after filtering
        is_3D = False

    # Filter spectrum
    spectrum_filt = np_lin.solve(Wmatf, spectrum[If, ...])

    if is_3D:
        if Nslice > 1:
            # Reshape to initial 3D shape
            spectrum_filt = np.reshape(
                spectrum_filt, (spectrum_filt.shape[0], Nangle, Nslice)
            )
        else:
            # Expand to 3D if requested
            spectrum_filt = spectrum_filt[:, :, None]

    if is_freq_pos:
        # Keep only positive frequencies
        Ip = freqs_th >= 0
        freqs_th = freqs_th[Ip]
        spectrum_filt = spectrum_filt[Ip, ...]

        # Multiply all components by two except zero frequency component
        spectrum_filt *= 2
        I0 = freqs_th == 0
        spectrum_filt[I0, ...] = 0.5 * spectrum_filt[I0, ...]

    if is_plot:
        import matplotlib.pyplot as plt

        # If_max = np.where(freqs > freqs_th.max())[0][0]
        # If_ref = np.arange(If_min, If_max, 1, dtype=int)
        # spec_val = np.abs(np.sqrt(np.sum(spectrum[If_ref, :] ** 2, axis=-1)))
        spec_val0 = np.abs(np.sqrt(np.sum(spectrum[If, :] ** 2, axis=-1)))
        # spec_val1 = np.abs(np.sqrt(np.sum(spectrum[If1, :] ** 2, axis=-1)))
        spec_val_filt = np.abs(np.sqrt(np.sum(spectrum_filt ** 2, axis=-1)))
        plt.figure()
        # plt.plot(freqs[If_ref], spec_val)
        plt.plot(freqs[If], spec_val0)
        # plt.plot(freqs[If1], spec_val1)
        plt.plot(freqs[If], spec_val_filt)
        plt.show()

    return spectrum_filt, freqs_th, Wmatf, If


def my_doorwin(f, Nt, dt, tol0=1e-4):
    """ "Return the Fourier transform of a door window as developed in to the method developed in
    Rainer et al., "Weak Coupling Between Electromagnetic and Structural Models for Electrical Machines",
    IEEE TRANSACTIONS ON MAGNETICS, VOL. 46, NO. 8, AUGUST 2010.

    Parameters
    ----------
    f : ndarray
        frequency array [Hz]
    Nt : int
        Number of time steps
    dt: float
        Time step value
    tol0: float
        absolute tolerance under which frequency is assumed to be 0

    Returns
    -------
    W : ndarray
        Door window Fourier transform

    """

    # Tolerance under which input frequency is considered as 0
    I0 = np.abs(f) > tol0

    W = np.ones([f.shape[0], f.shape[1]], dtype=complex)

    W[I0] = (
        (1 - np.exp(-1j * 2 * np.pi * dt * f[I0] * Nt))
        / (1 - np.exp(-1j * 2 * np.pi * dt * f[I0]))
        / Nt
    )

    return W



# -*- coding: utf-8 -*-

# File generated according to WVectorSelector.ui
# WARNING! All changes made in this file will be lost!
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_WVectorSelector(object):
    def setupUi(self, WVectorSelector):
        if not WVectorSelector.objectName():
            WVectorSelector.setObjectName("WVectorSelector")
        WVectorSelector.resize(218, 122)
        self.verticalLayout_2 = QVBoxLayout(WVectorSelector)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.g_vector_comp = QGroupBox(WVectorSelector)
        self.g_vector_comp.setObjectName("g_vector_comp")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.g_vector_comp.sizePolicy().hasHeightForWidth()
        )
        self.g_vector_comp.setSizePolicy(sizePolicy)
        self.g_vector_comp.setMinimumSize(QSize(0, 0))
        self.g_vector_comp.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout = QGridLayout(self.g_vector_comp)
        self.gridLayout.setObjectName("gridLayout")
        self.in_component = QLabel(self.g_vector_comp)
        self.in_component.setObjectName("in_component")
        self.in_component.setMinimumSize(QSize(0, 21))

        self.gridLayout.addWidget(self.in_component, 0, 0, 1, 1)

        self.c_component = QComboBox(self.g_vector_comp)
        self.c_component.addItem("")
        self.c_component.addItem("")
        self.c_component.addItem("")
        self.c_component.addItem("")
        self.c_component.addItem("")
        self.c_component.addItem("")
        self.c_component.setObjectName("c_component")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.c_component.sizePolicy().hasHeightForWidth())
        self.c_component.setSizePolicy(sizePolicy1)
        self.c_component.setMinimumSize(QSize(0, 21))
        self.c_component.setMaximumSize(QSize(16777215, 16777215))
        font = QFont()
        font.setBold(False)
        font.setWeight(50)
        self.c_component.setFont(font)

        self.gridLayout.addWidget(self.c_component, 0, 1, 1, 1)

        self.in_referential = QLabel(self.g_vector_comp)
        self.in_referential.setObjectName("in_referential")
        self.in_referential.setEnabled(True)
        self.in_referential.setMinimumSize(QSize(0, 21))

        self.gridLayout.addWidget(self.in_referential, 1, 0, 1, 1)

        self.c_referential = QComboBox(self.g_vector_comp)
        self.c_referential.addItem("")
        self.c_referential.addItem("")
        self.c_referential.setObjectName("c_referential")
        self.c_referential.setEnabled(True)
        sizePolicy1.setHeightForWidth(
            self.c_referential.sizePolicy().hasHeightForWidth()
        )
        self.c_referential.setSizePolicy(sizePolicy1)
        self.c_referential.setMinimumSize(QSize(100, 21))

        self.gridLayout.addWidget(self.c_referential, 1, 1, 1, 1)

        self.verticalLayout_2.addWidget(self.g_vector_comp)

        self.retranslateUi(WVectorSelector)

        QMetaObject.connectSlotsByName(WVectorSelector)

    # setupUi

    def retranslateUi(self, WVectorSelector):
        WVectorSelector.setWindowTitle(
            QCoreApplication.translate("WVectorSelector", "WVectorSelector", None)
        )
        self.g_vector_comp.setTitle("")
        self.in_component.setText(
            QCoreApplication.translate("WVectorSelector", "Component", None)
        )
        self.c_component.setItemText(
            0, QCoreApplication.translate("WVectorSelector", "radial", None)
        )
        self.c_component.setItemText(
            1, QCoreApplication.translate("WVectorSelector", "tangential", None)
        )
        self.c_component.setItemText(
            2, QCoreApplication.translate("WVectorSelector", "axial", None)
        )
        self.c_component.setItemText(
            3, QCoreApplication.translate("WVectorSelector", "comp_x", None)
        )
        self.c_component.setItemText(
            4, QCoreApplication.translate("WVectorSelector", "comp_y", None)
        )
        self.c_component.setItemText(
            5, QCoreApplication.translate("WVectorSelector", "comp_z", None)
        )

        self.in_referential.setText(
            QCoreApplication.translate("WVectorSelector", "Referential", None)
        )
        self.c_referential.setItemText(
            0, QCoreApplication.translate("WVectorSelector", "xyz", None)
        )
        self.c_referential.setItemText(
            1, QCoreApplication.translate("WVectorSelector", "radphiz", None)
        )

    # retranslateUi



from SciDataTool.Functions import AxisError


def check_filter(self):
    """Check if filter is correctly defined
    Parameters
    ----------
    self: Data1D
        a Data1D object
    Returns
    -------
    keys: list
        list of filter keys in same order as in values
    """
    keys = []
    for i, value in enumerate(self.values):
        items = value.split(self.delimiter)
        if len(items) != len(self.filter.keys()):
            raise AxisError("Filter is not correctly defined")
        for item in items:
            is_match = False
            for key in self.filter:
                if item in self.filter[key]:
                    is_match = True
                    if i == 0:
                        keys.append(key)
            if not is_match:
                raise AxisError("Missing item in filter: " + item)

    return keys



from PySide2.QtWidgets import QWidget
from PySide2.QtGui import QPixmap

from SciDataTool.GUI.WSliceOperator.Ui_WSliceOperator import Ui_WSliceOperator
from SciDataTool.GUI.WFilter.WFilter import WFilter
from PySide2.QtCore import Signal
from SciDataTool.Functions.Plot import axes_dict, fft_dict, ifft_dict, unit_dict
from SciDataTool.Classes.Data import Data
from numpy import where
from numpy import argmin, abs as np_abs
from os.path import join, abspath, dirname

type_extraction_dict = {
    "max": "=max",
    "min": "=min",
    "rms": "=rms",
    "rss": "=rss",
    "sum": "=sum",
    "mean": "=mean",
}

OPERATION_LIST = [
    "slice",
    "slice (fft)",
    "max",
    "min",
    "rms",
    "rss",
    "sum",
    "mean",
    "overlay",
]


class WSliceOperator(Ui_WSliceOperator, QWidget):
    """Widget to define how to handle the 'non-plot' axis"""

    refreshNeeded = Signal()
    generateAnimation = Signal()

    def __init__(self, parent=None, path_to_image=None):
        """Initialize the GUI according to info given by the WAxisManager widget

        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        parent : QWidget
            The parent QWidget
        path_to_image : str
            path to the folder where the image for the animation button is saved
        """

        # Build the interface according to the .ui file
        QWidget.__init__(self, parent=parent)
        self.setupUi(self)
        self.name = "angle"
        self.axis = Data
        self.current_dialog = None
        self.indices = None
        self.is_animate = False  # boolean to define if an animation must on this axis (must be on slice)
        self.is_pattern = False  # Detecting if the axis is a DataPattern (important for handling of slice + slider + lf_value)
        self.is_components = False
        self.path_to_image = path_to_image

        # Setting path to recover the image for the animate button
        if self.path_to_image is None:
            path_to_image = ":/images/images/icon/play-32px_SDT.png"
        else:
            path_to_image = self.path_to_image + "/images/icon/play-32px.png"

        self.b_animate.setPixmap(QPixmap(path_to_image))
        self.c_operation.currentTextChanged.connect(self.update_layout)
        self.slider.valueChanged.connect(self.update_floatEdit)
        self.slider.sliderReleased.connect(self.update_plot)
        self.lf_value.editingFinished.connect(self.update_slider)
        self.b_action.clicked.connect(self.open_filter)
        self.b_animate.mousePressEvent = self.gen_animate
        self.l_loading.setHidden(True)

    def gen_animate(self, e):
        """Methods called after clicking on animate button that emit a signal to generate a gif on the axis selected and display it
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object

        Output
        ---------
        None
        """
        self.is_animate = True
        self.generateAnimation.emit()

    def get_operation_selected(self):
        """Method that return a string of the action selected by the user on the axis of the widget.
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object

        Output
        ---------
        string
            name of the current action selected
        """

        # Recovering the action selected by the user
        action_type = self.c_operation.currentText().split(" over")[0]

        # Formatting the string to have the right syntax
        if action_type == "slice":
            # slice_index = self.slider.value()
            # action = "[" + str(slice_index) + "]"

            if self.is_pattern or self.is_components:
                # When working with DataPattern we must use index to give the exact value
                action = "[" + str(self.slider.value()) + "]"
            else:
                action = "=" + str(self.lf_value.value())

            return self.get_name() + action + "{" + self.unit + "}", self.is_animate

        elif action_type == "slice (fft)":
            action = "=" + str(self.lf_value.value())
            return self.get_name() + action, None

        elif action_type == "overlay":
            if self.indices is None:
                return self.get_name() + "[]", None
            else:
                return self.get_name() + str(self.indices), None

        elif action_type in type_extraction_dict:
            action = type_extraction_dict[action_type]
            return self.get_name() + action, None
        else:
            return None, None

    def get_name(self):
        """Method that return the name of the axis of the WSliceOperator
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        """
        return self.name

    def open_filter(self):
        """Open the Filter widget"""
        # Close previous dialog
        if self.current_dialog is not None:
            self.current_dialog.close()
            self.current_dialog.setParent(None)
            self.current_dialog = None

        self.current_dialog = WFilter(
            self.axis, self.indices, path_to_image=self.path_to_image
        )
        self.current_dialog.refreshNeeded.connect(self.update_indices)
        self.current_dialog.show()

    def update_indices(self):
        self.indices = self.current_dialog.indices
        self.refreshNeeded.emit()

    def set_name(self, name):
        """Method that set the name of the axis of the WSliceOperator
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        name : string
            string that hold the name of the axis
        """
        # Checking if the name of the axis is the name as the one displayed (z =/= axial direction for example)
        if name in axes_dict:
            self.in_name.setText(axes_dict[name])
        else:
            self.in_name.setText(name)

        self.name = name

    def set_operation(self, user_input):
        """Method that set the operation of the combobox of the WSliceOperator
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        user_input : list
            list of RequestedAxis that we use to set up the UI for the auto-plot
        """
        # Recovering the type of the operation and on which axis we are applying it
        operation_type = user_input.extension
        operation_name = user_input.name

        # Setting the label of the widget with the right name
        self.set_name(operation_name)

        # Converting type of the operation if we have a slice or an overlay
        if operation_type == "single":
            operation_type = "slice"

        elif operation_type == "list":
            operation_type = "overlay"

        else:
            operation_type += " over " + self.name

        # Setting operation combobox to the right operation
        self.c_operation.setCurrentIndex(self.c_operation.findText(operation_type))

        # Setting the slider to the right value if the operation is slice
        if operation_type == "slice":
            if user_input.indices is None:
                index = 0
            else:
                index = user_input.indices[0]
            self.set_slider(index)

    def set_slider(self, index):
        """Method that set the value of the slider of the WSliceOperator and then update floatEdit
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        index : int
            index at which the slider should be placed
        """
        self.slider.blockSignals(True)
        if index >= 0:
            self.slider.setValue(index)
        else:
            self.slider.setValue(self.slider.maximum() + index)
        self.slider.blockSignals(False)
        self.update_floatEdit()

    def set_slider_floatedit(self):
        """Method that set the value of the slider and the one of the floatEdit
        according to the axis sent by WAxisManager.
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        """
        if self.axis.is_components:
            self.is_components = True
            if self.c_operation.currentText() == "slice":
                self.axis_value = self.axis.values
            name = self.axis.name
        else:
            # Converting the axis from rad to degree if the axis is angle as we do slice in degrees
            # Recovering the value from the axis as well
            if self.c_operation.currentText() == "slice":
                if self.axis.name in ifft_dict:
                    operation = self.axis.name + "_to_" + self.axis_name
                else:
                    operation = None
                if self.axis_name in ifft_dict:
                    name = ifft_dict[self.axis_name]
                else:
                    name = self.axis_name
                self.set_name(name)
                if self.axis_name == "angle":
                    self.axis_value = self.axis.get_values(
                        unit="°", operation=operation, corr_unit="rad", is_full=True
                    )
                    self.unit = "°"
                else:
                    self.axis_value = self.axis.get_values(
                        operation=operation, is_full=True
                    )
            elif self.c_operation.currentText() == "slice (fft)":
                if self.axis_name in fft_dict:
                    name = fft_dict[self.axis_name]
                else:
                    name = self.axis_name
                self.set_name(name)
                if self.axis.name == "angle":
                    self.axis_value = self.axis.get_values(
                        operation="angle_to_wavenumber"
                    )
                    self.unit = ""
                elif self.axis.name == "time":
                    self.axis_value = self.axis.get_values(operation="time_to_freqs")
                    self.unit = "Hz"
                else:  # already wavenumber of freqs case
                    self.axis_value = self.axis.get_values()

        # Setting the axis unit
        if name in unit_dict:
            self.unit = unit_dict[name]

        if self.is_pattern:
            # Setting the initial value of the floatEdit to the minimum slice (=1)
            self.lf_value.setValue(1)
            self.in_unit.setText("[-]")
        elif self.is_components:
            self.lf_value.hide()
            self.b_animate.hide()
            self.in_unit.setText(self.axis_value[0])
        else:
            # Setting the initial value of the floatEdit to the minimum inside the axis
            self.lf_value.setValue(min(self.axis_value))
            self.in_unit.setText("[" + self.unit + "]")

        # Setting the slider by giving the number of index according to the size of the axis
        self.slider.setMinimum(0)
        self.slider.setMaximum(len(self.axis_value) - 1)
        self.slider.setValue(0)

    def update(self, axis, axis_request=None):
        """Method that will update the WSliceOperator widget according to the axis given to it
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        axis : string
            string with the name of the axis that should set the WSliceOperator widget
        """
        self.axis = axis

        # Relative import of DataPattern to prevent circular import
        module = __import__("SciDataTool.Classes.DataPattern", fromlist=["DataPattern"])
        DataPattern = getattr(module, "DataPattern")

        # Detecting if the axis is a DataPattern (important for handling of slice + slider + lf_value)
        if isinstance(self.axis, DataPattern):
            self.is_pattern = True

        if axis_request is not None:
            axis_request.corr_unit = axis_request.unit
            axis_request.get_axis(axis, True)
            self.indices = axis_request.indices
        self.unit = axis.unit
        if axis.name in ifft_dict:  # DataFreq case
            self.axis_name = ifft_dict[axis.name]
        else:
            self.axis_name = axis.name
        self.set_name(self.axis_name)

        self.c_operation.blockSignals(True)
        operation_list = [
            ope + " over " + self.name if ope in type_extraction_dict else ope
            for ope in OPERATION_LIST
        ]

        # Remove slice for string axes
        if self.axis.is_components:
            operation_list.remove("slice")
        else:
            self.set_slider_floatedit()

        # Remove overlay for non is_components axes
        if not self.axis.is_components:
            operation_list.remove("overlay")

        # Remove fft slice for non fft axes
        if not self.axis_name in fft_dict:
            operation_list.remove("slice (fft)")

        self.c_operation.clear()
        self.c_operation.addItems(operation_list)
        self.update_layout()
        if self.axis.is_overlay:
            self.c_operation.setCurrentIndex(operation_list.index("overlay"))
            self.b_action.show()
            self.b_action.setText("Overlay")
        self.c_operation.blockSignals(False)

    def update_floatEdit(self):
        """Method that set the value of the floatEdit according to the value returned by the slider
        and the axis sent by WAxisManager.
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        """

        self.lf_value.blockSignals(True)

        if self.is_pattern:
            # When working with DataPattern, we have to work with the exact value given
            self.lf_value.setValue(self.slider.value() + 1)
        elif self.is_components:
            self.in_unit.setText(self.axis_value[self.slider.value()])
        else:
            self.lf_value.setValue(self.axis_value[self.slider.value()])

        self.lf_value.blockSignals(False)

    def update_plot(self):
        """Method that set the value of the floatEdit according to the value returned by the slider
        and the axis sent by WAxisManager.
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        """

        self.lf_value.blockSignals(True)

        if self.is_pattern:
            # When working with DataPattern, we have to work with the exact value given
            self.lf_value.setValue(self.slider.value() + 1)
        elif self.is_components:
            self.in_unit.setText(self.axis_value[self.slider.value()])
        else:
            self.lf_value.setValue(self.axis_value[self.slider.value()])

        self.lf_value.blockSignals(False)
        self.refreshNeeded.emit()

    def update_layout(self):
        """Method that update the layout of the WSliceOperator according to the extraction chosen
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        """
        # Recovering the operation selected
        extraction_selected = self.c_operation.currentText().split(" over")[0]

        # If the operation selected is a slice, then we show the slider and the floatEdit
        if extraction_selected == "slice" or extraction_selected == "slice (fft)":
            self.set_slider_floatedit()
            if not self.is_components:
                self.lf_value.show()
            self.in_unit.show()
            self.slider.show()
            self.b_action.hide()
            if extraction_selected == "slice" and not self.is_components:
                self.b_animate.show()
            else:
                self.b_animate.hide()
            self.refreshNeeded.emit()
        # If the operation selected is overlay then we show the related button
        elif extraction_selected == "overlay":
            if self.axis_name in ifft_dict:
                name = ifft_dict[self.axis_name]
            else:
                name = self.axis_name
            self.set_name(name)
            self.lf_value.hide()
            self.in_unit.hide()
            self.slider.hide()
            self.b_animate.hide()
            self.b_action.show()
            self.b_action.setText("Overlay")
            self.refreshNeeded.emit()
        else:
            if self.axis_name in ifft_dict:
                name = ifft_dict[self.axis_name]
            else:
                name = self.axis_name
            self.set_name(name)
            self.lf_value.hide()
            self.in_unit.hide()
            self.slider.hide()
            self.b_animate.hide()
            self.b_action.hide()
            self.refreshNeeded.emit()

    def update_slider(self):
        """Method that set the value of the slider according to the value of the floatEdit
        according to the axis sent by WAxisManager.
        Parameters
        ----------
        self : WSliceOperator
            a WSliceOperator object
        """

        self.slider.blockSignals(True)
        # We set the value of the slider to the index closest to the value given
        if self.is_pattern:
            # When working with DataPattern, we have to work with the exact value given
            self.slider.setValue(self.lf_value.value() - 1)
        else:
            index = argmin(np_abs(self.axis_value - self.lf_value.value()))
            self.slider.setValue(index)
        # We update the value of floatEdit to the index selected
        # self.lf_value.setValue(self.axis_value[index])
        self.slider.blockSignals(False)
        self.refreshNeeded.emit()



from PySide2.QtCore import Qt
from PySide2.QtWidgets import QCheckBox


class CheckBox(QCheckBox):
    """To bypass partiallychecked state with user"""

    def __init__(self, *args, **kwargs):
        """Same constructor as QCheckBox"""
        # Call the QCheckBox constructor
        super(CheckBox, self).__init__(*args, **kwargs)
        self.setTristate(True)

    def nextCheckState(self):
        if self.checkState() == Qt.Unchecked:
            super(CheckBox, self).nextCheckState()  # partiallychecked
            super(CheckBox, self).nextCheckState()  # checked
        else:
            super(CheckBox, self).nextCheckState()



from importlib import import_module
from numpy import array

from SciDataTool.Functions.conversions import convert
from SciDataTool.Functions import NormError


def get_values(
    self,
    unit="SI",
    is_oneperiod=False,
    is_antiperiod=False,
    is_smallestperiod=False,
    is_pattern=False,
    normalization=None,
    operation=None,
    is_real=True,
    corr_unit=None,
    is_full=False,
):
    """Returns the vector 'axis' taking symmetries into account.
    Parameters
    ----------
    self: DataPattern
        a DataPattern object
    unit: str
        requested unit
    is_oneperiod: bool
        return values on a single period (unused here)
    is_antiperiod: bool
        return values on a semi period (unused here)
    is_smallestperiod: bool
        return values on smallest available period (pattern here => is_pattern)
    is_pattern: bool
        return values on smallest available pattern
    normalization: str
        name of normalization to use
    operation: str
        name of the operation (e.g. "freqs_to_time")
    Returns
    -------
    Vector of axis values
    """
    values = self.values

    # Rebuild pattern
    if is_smallestperiod or is_pattern:
        pass
    else:
        if self.values_whole is not None:
            values = self.values_whole
        else:
            values = values[self.rebuild_indices]

    # fft/ifft
    if operation is not None:
        module = import_module("SciDataTool.Functions.conversions")
        func = getattr(module, operation)  # Conversion function
        values = array(func(values, is_real=is_real))

    if unit != "SI" and unit != self.unit:
        if unit in self.normalizations and normalization is None:
            normalization = unit
            unit = "SI"

    # Normalization
    if normalization is not None:
        if normalization in self.normalizations:
            if (
                self.normalizations[normalization].unit == "SI"
                or self.normalizations[normalization].unit == self.unit
            ):
                # Axis is in the correct unit for the normalization
                values = self.normalizations[normalization].normalize(values)
            else:
                raise NormError("Normalization is not available in this unit")
        else:
            raise NormError("Normalization is not available")

    # Unit conversion
    if unit != "SI" and unit != self.unit:
        if corr_unit is not None:
            values = convert(values, corr_unit, unit)
        else:
            values = convert(values, self.unit, unit)

    return values



from SciDataTool.Functions.Plot.plot_2D import plot_2D
from SciDataTool.Functions.Plot import (
    unit_dict,
    norm_dict,
    axes_dict,
    fft_dict,
    COLORS,
)
from SciDataTool.Functions.Load.import_class import import_class
from SciDataTool.Functions.fix_axes_order import fix_axes_order
from SciDataTool.Functions.parser import read_input_strings
from SciDataTool.Classes.Norm_indices import Norm_indices
from numpy import (
    squeeze,
    split,
    array,
    where,
    unique,
    nanmax as np_max,
    array2string,
    insert,
    nanmin as np_min,
    linspace,
    argmin,
    argmax,
    take,
    log10,
    nan,
)


def plot_2D_Data(
    self,
    *arg_list,
    axis_data=None,
    is_norm=False,
    unit="SI",
    overall_axes=[],
    data_list=[],
    legend_list=[],
    color_list=None,
    linestyles=None,
    linewidth_list=[2],
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_disp_title=True,
    is_grid=True,
    is_auto_ticks=True,
    is_auto_range=True,
    xlabel=None,
    ylabel=None,
    title=None,
    fig=None,
    ax=None,
    barwidth=100,
    type_plot=None,
    fund_harm_dict=None,
    is_show_fig=None,
    win_title=None,
    thresh=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
    is_show_legend=True,
    is_outside_legend=False,
    is_frame_legend=True,
    is_indlabels=False,
    annotations=None,
):
    """Plots a field as a function of time

    Parameters
    ----------
    data : Data
        a Data object
    *arg_list : list of str
        arguments to specify which axes to plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    data_list : list
        list of Data objects to compare
    legend_list : list
        list of legends to use for each Data object (including reference one) instead of data.name
    color_list : list
        list of colors to use for each Data object
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    is_grid : bool
        boolean indicating if the grid must be displayed
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    barwidth : float
        barwidth scaling factor, only if type_plot = "bargraph"
    type_plot : str
        type of 2D graph : "curve", "bargraph", "barchart" or "quiver"
    fund_harm_dict : dict
        Dict containing axis name as key and frequency/order/wavenumber of fundamental harmonic as value to display fundamental harmonic in red in the fft
    is_show_fig : bool
        True to show figure after plot
    win_title : str
        Title of the plot window
    thresh : float
        threshold for automatic fft ticks
    is_outside_legend : bool
        True to display legend outside the graph
    is_frame_legend : bool
        True to display legend in a frame
    """

    # Dynamic import to avoid import loop
    DataPattern = import_class("SciDataTool.Classes", "DataPattern")

    # Extract arg_list it the function called from another script with *arg_list
    if len(arg_list) == 1 and type(arg_list[0]) == tuple:
        arg_list = arg_list[0]

    axes_names = [
        axis.name for axis in read_input_strings(arg_list, axis_data=axis_data)
    ]

    # Fix axes order
    arg_list_along = fix_axes_order([axis.name for axis in self.get_axes()], arg_list)

    # In case of 1D fft, keep only positive wavenumbers
    for i, arg in enumerate(arg_list_along):
        if "wavenumber" in arg and "=" not in arg and "[" not in arg:
            liste = list(arg_list_along)
            liste[i] = arg.replace("wavenumber", "wavenumber>0")
            arg_list_along = tuple(liste)

    if color_list == [] or color_list is None:
        color_list = COLORS

    new_color_list = color_list.copy()

    # Set unit
    if unit == "SI":
        unit = self.unit

    # Detect if is fft, build ylabel
    if "dB" in unit and "ref" in self.normalizations:
        unit_str = (
            "["
            + unit
            + " re. "
            + str(self.normalizations["ref"].ref)
            + " $"
            + self.unit
            + "$]"
        )
    else:
        unit_str = r"$[" + unit + "]$"
    is_fft = False
    if (
        any("wavenumber" in s for s in arg_list) or any("freqs" in s for s in arg_list)
    ) and type_plot != "curve":
        is_fft = True
        if self.symbol == "Magnitude":
            if ylabel is None:
                ylabel = "Magnitude " + unit_str
        else:
            if ylabel is None:
                ylabel = r"$|\widehat{" + self.symbol + "}|$ " + unit_str
    else:
        if is_norm:
            if ylabel is None:
                ylabel = (
                    r"$\frac{" + self.symbol + "}{" + self.symbol + "_0}\,$" + unit_str
                )
        else:
            if self.symbol == "Magnitude":
                if ylabel is None:
                    ylabel = "Magnitude " + unit_str
            else:
                if ylabel is None:
                    ylabel = r"$" + self.symbol + "\,$" + unit_str

    # Extract field and axes
    Xdatas = []
    Ydatas = []
    data_list2 = [self] + data_list
    for i, d in enumerate(data_list2):
        if is_fft or "dB" in unit:
            result = d.get_magnitude_along(
                *arg_list_along, axis_data=axis_data, unit=unit, is_norm=is_norm
            )
            if i == 0:
                axes_list = result.pop("axes_list")
                axes_dict_other = result.pop("axes_dict_other")
                result_0 = result
        else:
            result = d.get_along(
                *arg_list_along, axis_data=axis_data, unit=unit, is_norm=is_norm
            )
            if i == 0:
                axes_list = result.pop("axes_list")
                axes_dict_other = result.pop("axes_dict_other")
                result_0 = result
        Ydatas.append(result.pop(d.symbol))
        # in string case not overlay, Xdatas is a linspace
        if (
            axes_list[list(result.keys()).index(axes_names[0])].is_components
            and axes_list[list(result.keys()).index(axes_names[0])].extension != "list"
        ):
            xdata = linspace(
                0, len(result[axes_names[0]]) - 1, len(result[axes_names[0]])
            )
        else:
            xdata = result[axes_names[0]]
        Xdatas.append(xdata)

    # Build xlabel and title
    title1 = self.name[0].capitalize() + self.name[1:] + " "
    title2 = "for "
    for axis in axes_list:
        if axis.unit in norm_dict and axis.unit != "Hz":
            name = norm_dict[axis.unit].split(" [")[0]
        elif axis.name in axes_dict:
            name = axes_dict[axis.name]
        else:
            name = axis.name
        if (
            axis.extension
            in [
                "whole",
                "interval",
                "oneperiod",
                "antiperiod",
                "smallestperiod",
                "axis_data",
                "list",
            ]
            and len(axis.values) > 1
            or (len(axis.values) == 1 and len(axes_list) == 1)
        ):
            if axis.unit == "SI":
                if axis.name in unit_dict:
                    axis_unit = unit_dict[axis.name]
                else:
                    axis_unit = axis.unit
                if xlabel is None:
                    xlabel = name[0].capitalize() + name[1:] + " [" + axis_unit + "]"
                main_axis_name = name
            elif axis.unit in norm_dict:
                if xlabel is None:
                    xlabel = norm_dict[axis.unit]
                if axis.unit == "Hz":
                    main_axis_name = "frequency"
                else:
                    main_axis_name = axis.unit
            else:
                axis_unit = axis.unit
                if xlabel is None:
                    xlabel = name[0].capitalize() + name[1:] + " [" + axis_unit + "]"
                main_axis_name = name
            if (
                axis.name == "angle"
                and axis.unit == "°"
                and round(np_max(axis.values) / 6) % 5 == 0
            ):
                xticks = [i * round(np_max(axis.values) / 6) for i in range(7)]
            else:
                xticks = None
            if (
                axes_list[list(result.keys()).index(axes_names[0])].is_components
                and not self.get_axes()[0].is_overlay
            ):
                xticklabels = result[axes_names[0]]
                xticks = Xdatas[0]
            else:
                xticklabels = None
                xticks = None
        else:
            if axis.corr_unit == "SI":
                if axis.name in unit_dict:
                    axis_unit = unit_dict[axis.name]
                else:
                    axis_unit = axis.unit
            elif axis.corr_unit in norm_dict and axis.corr_unit != "Hz":
                axis_unit = norm_dict[axis.corr_unit]
            else:
                axis_unit = axis.unit

            if isinstance(result_0[axis.name], str):
                title2 += name + "=" + result_0[axis.name]
            else:
                if isinstance(result_0[axis.name][0], str):
                    axis_str = result_0[axis.name][0]
                else:
                    if result_0[axis.name][0] > 10:
                        fmt = "{:.5g}"
                    else:
                        fmt = "{:.3g}"
                    axis_str = array2string(
                        result_0[axis.name], formatter={"float_kind": fmt.format}
                    ).replace(" ", ", ")

                if len(result_0[axis.name]) == 1:
                    axis_str = axis_str.strip("[]")

                index = None
                if axis.is_pattern and len(axis.values) == 1:
                    for axis_obj in self.get_axes():
                        if axis_obj.name == axis.name:
                            index = axis.indices[0]

                title2 += name + "=" + axis_str.rstrip(", ") + " [" + axis_unit + "]"
                if index is not None:
                    title2 += " (slice " + str(index + 1) + "), "
                else:
                    title2 += ", "

    # Title part 3 containing axes that are here but not involved in requested axes
    title3 = ""
    for axis_name in axes_dict_other:
        is_display = True
        for axis in self.axes:
            if axis.name == axis_name:
                if isinstance(axis, DataPattern) and len(axis.unique_indices) == 1:
                    is_display = False
        if is_display:
            if isinstance(axes_dict_other[axis_name][0], str):
                title3 += axis_name + "=" + axes_dict_other[axis_name][0]
            else:
                if axes_dict_other[axis_name][0] > 10:
                    fmt = "{:.5g}"
                else:
                    fmt = "{:.3g}"
                title3 += (
                    axis_name
                    + "="
                    + array2string(
                        axes_dict_other[axis_name][0],
                        formatter={"float_kind": fmt.format},
                    ).replace(" ", ", ")
                    + " ["
                    + axes_dict_other[axis_name][1]
                    + "], "
                )

    if title2 == "for " and title3 == "":
        title2 = ""

    # Detect discontinuous axis (Norm_indices) to use bargraph
    for axis in axes_list:
        if axis.unit in self.axes[axis.index].normalizations:
            if isinstance(
                self.axes[axis.index].normalizations[axis.unit], Norm_indices
            ):
                type_plot = "bargraph"

    # Detect how many curves are overlaid, build legend and color lists
    if legend_list == [] and data_list != []:
        legend_list = [d.name for d in data_list2]
    elif legend_list == []:
        legend_list = ["" for d in data_list2]
    legends = []
    # Prepare colors
    linestyle_list = linestyles
    for i, d in enumerate(data_list2):
        is_overlay = False
        for axis in axes_list:
            if axis.extension == "list":
                is_overlay = True
                if linestyles is None:
                    linestyles = ["dashed"]
                n_curves = len(axis.values)
                if axis.unit == "SI":
                    if axis.name in unit_dict:
                        axis_unit = unit_dict[axis.name]
                    else:
                        axis_unit = axis.unit
                elif axis.unit in norm_dict:
                    axis_unit = norm_dict[axis.unit]
                else:
                    axis_unit = axis.unit
                if len(d.axes[axis.index].get_values()) > 1:
                    legends += [
                        (
                            legend_list[i]
                            + " "
                            + axis.name
                            + "="
                            + axis.values.tolist()[j]
                            + " "
                            + axis_unit
                            if isinstance(axis.values.tolist()[j], str)
                            else legend_list[i]
                            + " "
                            + axis.name
                            + "="
                            + "%.3g" % axis.values.tolist()[j]
                            + " "
                            + axis_unit
                        )
                        .replace("SI", "")
                        .replace(" []", "")
                        for j in range(n_curves)
                    ]
                else:
                    legends += [legend_list[i]]

        if not is_overlay:
            legends += [legend_list[i]]
            # Adjust colors in non overlay case with overlay axis
            if len(data_list2) > 1:
                for axis in self.get_axes():
                    if axis.is_overlay and len(color_list) > len(axis.values):
                        new_color_list[1:] = color_list[len(axis.values) :]

    # Split Ydatas if the plot overlays several curves
    if is_overlay:
        Ydata = []
        for d in Ydatas:
            if d.ndim != 1:
                axis_index = where(array(d.shape) == n_curves)[0]
                if axis_index.size > 1:
                    print("WARNING, several axes with same dimensions")
                Ydata += split(d, n_curves, axis=axis_index[-1])
            else:
                Ydata += [d]
        Ydatas = [squeeze(d) for d in Ydata]
        Xdata = []
        for i in range(len(data_list2)):
            Xdata += [Xdatas[i] for x in range(n_curves)]
        Xdatas = Xdata

    # Finish title
    if title is None:

        # Reformat in case of operation
        for ope in ["max", "min", "mean"]:
            if "=" + ope in title2:
                title2 = ""
                title1 = (
                    ope[0].capitalize() + ope[1:] + " " + title1[0].lower() + title1[1:]
                )

        # Concatenate all title parts
        if is_overlay:
            title = title1 + title3
        else:
            title = title1 + title2 + title3

        # Remove last coma due to title2 or title3
        title = title.rstrip(", ")

        # Remove dimless and quotes
        title = title.replace("SI", "").replace("[]", "").replace("'", "")

    xlabel = xlabel.replace("SI", "").replace(" []", "")  # Remove dimless units
    ylabel = ylabel.replace("SI", "").replace(" []", "")  # Remove dimless units

    # Overall computation
    if overall_axes != []:
        if self.unit == "W":
            op = "=sum"
        else:
            op = "=rss"
        arg_list_ovl = [0 for i in range(len(arg_list))]
        # Add sum to overall_axes
        for axis in overall_axes:
            is_match = False
            for i, arg in enumerate(arg_list):
                if axis in arg:
                    is_match = True
                    arg_list_ovl[i] = axis + op
            if not is_match:
                arg_list_ovl.append(axis + op)
        # Add other requested axes
        for i, arg in enumerate(arg_list):
            if arg_list_ovl[i] == 0:
                arg_list_ovl[i] = arg
        if is_fft or "dB" in unit:
            result = self.get_magnitude_along(
                *arg_list_ovl, unit=unit, axis_data=axis_data
            )
        else:
            result = self.get_along(*arg_list_ovl, unit=unit, axis_data=axis_data)
        Y_overall = result[self.symbol]
        # in string case not overlay, Xdatas is a linspace
        if (
            axes_list[list(result.keys()).index(axes_names[0])].is_components
            and axes_list[list(result.keys()).index(axes_names[0])].extension != "list"
        ):
            xdata = linspace(
                0, len(result[axes_names[0]]) - 1, len(result[axes_names[0]])
            )
        else:
            xdata = result[list(result)[0]]
        Ydatas.insert(0, Y_overall)
        Xdatas.insert(0, xdata)
        new_color_list = new_color_list.copy()
        new_color_list.insert(0, "#000000")
        legends.insert(0, "Overall")

    # Deactivate legend if only one item
    if type_plot == "point":
        legends = []

    if "dB" in unit:  # Replace <=0 by nans
        for ydata in Ydatas:
            ydata[ydata <= 0] = nan

    # Call generic plot function
    if is_fft:

        if thresh is None:
            if self.normalizations is not None and "ref" in self.normalizations:
                thresh = self.normalizations["ref"].ref
            else:
                thresh = 0.02

        freqs = Xdatas[0]
        if "dB" in unit:
            indices = [
                ind
                for ind, y in enumerate(Ydatas[0])
                if abs(y) > max(10 * log10(thresh) + abs(np_max(Ydatas[0])), 0)
            ]
        else:
            if Ydatas[0].size == 1:
                indices = [0]
            else:
                indices = [
                    ind
                    for ind, y in enumerate(Ydatas[0])
                    if abs(y) > abs(thresh * np_max(Ydatas[0]))
                ]
        xticks = unique(insert(freqs[indices], 0, 0))

        if is_auto_range:
            if len(xticks) > 1:
                if x_min is None:
                    x_min = xticks[0]
                else:
                    x_min = max(x_min, xticks[0])
                if x_max is None:
                    x_max = xticks[-1]
                else:
                    x_max = min(x_max, xticks[-1])
            else:
                if x_min is None:
                    x_min = np_min(freqs)
                else:
                    x_min = max(x_min, np_min(freqs))
                if x_max is None:
                    x_max = np_max(freqs)
                else:
                    x_max = min(x_max, np_max(freqs))

        else:
            if x_min is None:
                x_min = np_min(freqs)
            if x_max is None:
                x_max = np_max(freqs)

        x_min = x_min - x_max * 0.05
        x_max = x_max * 1.05

        if (
            len(xticks) == 0
            or (
                len(xticks) > 20
                and not axes_list[
                    list(result.keys()).index(axes_names[0])
                ].is_components
            )
            or not is_auto_ticks
        ):
            xticks = None

        # Force bargraph for fft if type_graph not specified
        if type_plot is None:
            type_plot = "bargraph"

        # Option to draw fundamental harmonic in red
        if not fund_harm_dict:
            fund_harm = None
        else:
            # Activate the option only if main axis is in dict and only one Data is plotted
            if main_axis_name in fund_harm_dict and len(Ydatas) == 1:
                fund_harm = fund_harm_dict[main_axis_name]
            else:
                # Deactivate the option
                fund_harm = None

        plot_2D(
            Xdatas,
            Ydatas,
            legend_list=legends,
            color_list=new_color_list,
            linestyle_list=linestyle_list,
            linewidth_list=linewidth_list,
            fig=fig,
            ax=ax,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            type_plot=type_plot,
            x_min=x_min,
            x_max=x_max,
            y_min=y_min,
            y_max=y_max,
            is_logscale_x=is_logscale_x,
            is_logscale_y=is_logscale_y,
            is_disp_title=is_disp_title,
            is_grid=is_grid,
            xticks=xticks,
            xticklabels=xticklabels,
            save_path=save_path,
            barwidth=barwidth,
            fund_harm=fund_harm,
            is_show_fig=is_show_fig,
            win_title=win_title,
            font_name=font_name,
            font_size_title=font_size_title,
            font_size_label=font_size_label,
            font_size_legend=font_size_legend,
            is_show_legend=is_show_legend,
            is_outside_legend=is_outside_legend,
            is_frame_legend=is_frame_legend,
            is_indlabels=is_indlabels,
        )

    else:

        # Force curve plot if type_plot not specified
        if type_plot is None:
            type_plot = "curve"
        annot = None
        # Hidden annotations
        if annotations is not None:
            try:
                annot = list()
                axis_along = self.get_axes(annotations[0])[0]
                axis_op = self.get_axes(annotations[1])[0]
                operation = annotations[2]
                arg_list_new = []
                if self.unit == "W":
                    op = "=sum"
                else:
                    op = "=rss"
                for axis in self.get_axes():
                    if axis.name not in [axis_along.name, axis_op.name]:
                        arg_list_new.append(axis.name + op)
                    else:
                        arg_list_new.append(axis.name)
                data2 = self.get_data_along(*arg_list_new, unit=unit)
                arg_list_new = []
                for arg in arg_list_along:
                    if axis_along.name in arg or axis_op.name in arg:
                        arg_list_new.append(arg.replace("=sum", ""))
                result = data2.get_magnitude_along(*arg_list_new)
                for ii in range(len(result[annotations[0]])):
                    if operation == "max":
                        index = argmax(take(result[self.symbol], ii, axis=0))
                        annot.append(
                            "main "
                            + annotations[1].rstrip("s")
                            + ": "
                            + axis_op.get_values()[index]
                        )
                if len(overall_axes) != 0:
                    annot.insert(0, "Overall")
            except Exception:
                pass

        plot_2D(
            Xdatas,
            Ydatas,
            legend_list=legends,
            color_list=new_color_list,
            fig=fig,
            ax=ax,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            type_plot=type_plot,
            x_min=x_min,
            x_max=x_max,
            y_min=y_min,
            y_max=y_max,
            is_logscale_x=is_logscale_x,
            is_logscale_y=is_logscale_y,
            is_disp_title=is_disp_title,
            is_grid=is_grid,
            xticks=xticks,
            xticklabels=xticklabels,
            barwidth=barwidth,
            linestyle_list=linestyle_list,
            linewidth_list=linewidth_list,
            save_path=save_path,
            is_show_fig=is_show_fig,
            win_title=win_title,
            font_name=font_name,
            font_size_title=font_size_title,
            font_size_label=font_size_label,
            font_size_legend=font_size_legend,
            is_show_legend=is_show_legend,
            is_outside_legend=is_outside_legend,
            is_frame_legend=is_frame_legend,
            is_indlabels=is_indlabels,
            annotations=annot,
        )



import numpy as np

from SciDataTool.Functions.set_routines import unique_tol


def get_conv_indices(freqs1, freqs2, f_min, f_max, tol_freq):
    """Compute frequency array and return indices to perform DataFreq convolution

    Parameters
    ----------
    freqs1 : ndarray
        frequency array of first DataFreq
    freqs2 : ndarray
        frequency array of second DataFreq
    f_min: Float
        Minimum frequency of convolved spectrum [Hz]
    f_max: Float
        Maximum frequency of convolved spectrum [Hz]
    tol_freq: Float
        Absolute tolerance value to filter harmonic orders by their frequency value

    Returns
    -------
    freqs_un : ndarray
        frequency array of the convolved spectrum
    I0a: ndarray
        array of indices to extract only amplitudes which are kept for convolution (positive half of the spectrum)
    I0b: ndarray
        True = to get array of positive half of spectrum
        False = to get array of negative part of spectrum
    I1: ndarray
        array on which amplitudes are accumulated
    I2: ndarray
        index of constant component
    """

    # In case f_min is zero, consider also negative frequencies larger than tol_freq
    if f_min in [0, None]:
        f_min = -tol_freq

    # Get harmonic numbers in both spectrums
    Nh1 = freqs1.size
    Nh2 = freqs2.size

    # Compute the order array containing all combinations of input frequencies
    # Mirror n1 and n2 with their negative parts
    f1_0 = np.concatenate((freqs1, -freqs1), axis=0)
    f2_0 = np.concatenate((freqs2, -freqs2), axis=0)

    # Calculate meshgrid
    f1_1, f2_1 = np.meshgrid(f1_0, f2_0)

    # Calculate all frequency combinations and flatten grids
    freqs = f1_1.ravel("C") + f2_1.ravel("C")

    # Restrict convolution to f_min, f_max
    I0 = np.logical_and.reduce((freqs >= f_min, freqs <= f_max))

    # Keep only unique rows in order array

    freqs_un, I1 = unique_tol(
        freqs[I0],
        return_inverse=True,
        return_index=False,
        axis=0,
        is_abs_tol=True,
        tol=tol_freq,
    )

    # Find indices of positive frequencies to be kept
    I0a = np.where(I0)[0]

    if np.any(np.abs(freqs1) < tol_freq) and np.any(np.abs(freqs2) < tol_freq):
        # Convolution product cannot be optimized using conjugate property
        I0b = np.array([])
    else:
        # Find indices of components that will be conjugated
        I0b = I0a + 1 <= 2 * Nh1 * Nh2

        # Shift indices of the conjugated part to link them to their conjugate
        I0a[~I0b] = I0a[~I0b] + Nh1 - 2 * Nh1 * Nh2

    # Find f=0 component to divide by two after convolution
    I2 = np.where(freqs_un < tol_freq)[0]

    return freqs_un, I0a, I0b, I1, I2


def get_conv_amplitudes(amp1, amp2, I0a, I0b, I1, I2):
    """Compute amplitudes resulting from the convolution of input values

    Parameters
    ----------
    amp1: ndarray
        amplitude of first spectrum
    amp2: ndarray
        amplitude of second spectrum
    I0a: ndarray
        array of positive half of spectrum
    I0b: ndarray
        array of negative part of spectrum
    I1: ndarray
        array on which amplitudes are accumulated
    I2: ndarray
        index of constant component

    Returns
    -------
    amp: ndarray
        array of harmonics amplitude
    """

    # Duplicate and conjugate complex magnitudes to account for negative frequencies (1st half)
    amp1_1 = np.concatenate((amp1, np.conjugate(amp1)), axis=0)

    if I0b.size > 0:
        # Use conjugate property to reduce outer product size
        # Outer product of both complex amplitude vectors to calculte all harmonic combinations
        amp3 = np.outer(amp1_1, amp2).ravel("F")

        # Duplicate and conjugate complex magnitudes to account for negative frequencies (2nd half)
        amp_full = (
            np.concatenate((amp3[I0a[I0b]], np.conjugate(amp3[I0a[~I0b]])), axis=0) / 2
        )

    else:
        # Duplicate and conjugate complex magnitudes to account for negative frequencies
        amp2_2 = np.concatenate((amp2, np.conjugate(amp2)), axis=0)

        # Outer product of both complex amplitude vectors to calculte all harmonic combinations
        amp_full = np.outer(amp1_1, amp2_2).ravel("F")[I0a] / 2

    # Sum all contributions which have the same index
    if amp_full.dtype == complex:
        amp = np.bincount(I1, weights=amp_full.real) + 1j * np.bincount(
            I1, weights=amp_full.imag
        )
    else:
        amp = np.bincount(I1, weights=amp_full)

    # Divide by two constant component
    if I2.size > 0:
        amp[I2] /= 2

    return amp


def get_sum_indices(freqs1, freqs2, tol_freq):
    """Compute frequency array and return indices to perform DataFreq sum

    Parameters
    ----------
    freqs1 : ndarray
        frequency array of first DataFreq
    freqs2 : ndarray
        frequency array of second DataFreq
    tol_freq: Float
        Absolute tolerance value to filter harmonic orders by their frequency value

    Returns
    -------
    freqs_un : ndarray
        frequency array of summed DataFreq
    I0b: ndarray
        array on which amplitudes are accumulated
    """

    # Compute the frequency array containing all combinations
    freqs = np.concatenate((freqs1, freqs2), axis=0)

    # Get unique orders
    freqs_un, I0b = unique_tol(
        freqs,
        return_inverse=True,
        return_index=False,
        axis=0,
        is_abs_tol=True,
        tol=tol_freq,
    )

    return freqs_un, I0b


def get_sum_amplitudes(amp1, amp2, I0b):
    """Calculate summed amplitudes of both input amplitudes arrays

    Parameters
    ----------
    amp1 : ndarray
        First amplitude array
    amp2 : ndarray
        Second amplitude array
    I0b: ndarray
        array on which amplitudes are accumulated

    Returns
    -------
    amp: ndarray
        array of harmonics amplitude
    """

    # Concatenate both amplitude arrays
    amp_full = np.concatenate((amp1, amp2), axis=0)

    # Sum all contributions which have the same orders and wavenumber
    if amp_full.dtype == complex:
        amp = np.bincount(I0b, weights=amp_full.real) + 1j * np.bincount(
            I0b, weights=amp_full.imag
        )
    else:
        amp = np.bincount(I0b, weights=amp_full)

    return amp



def to_datadual(self, datafreq=None):
    """Creates a DataDual object.
    Parameters
    ----------
    self : DataTime
        a DataTime object
    datafreq : DataFreq
        a DataFreq object
    Returns
    -------
    a DataDual object
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.DataDual", fromlist=["DataDual"])
    DataDual = getattr(module, "DataDual")

    # Fill DataDual object with time/space data
    datadual = DataDual(
        name=self.name,
        symbol=self.symbol,
        unit=self.unit,
        normalizations=self.normalizations.copy(),
        axes_dt=self.axes,
        values_dt=self.values,
    )

    # Add fourier data (from input or using fft)
    if datafreq is None:
        datafreq = self.time_to_freq()
    datadual.axes_df = datafreq.axes
    datadual.values_df = datafreq.values

    return datadual



from numpy import array, argsort

from importlib import import_module

from SciDataTool.Classes.DataPattern import DataPattern

from SciDataTool.Functions.symmetries import rebuild_symmetries_axis
from SciDataTool.Functions import AxisError

operation_list = [
    "max",
    "min",
    "sum",
    "rss",
    "mean",
    "rms",
    "integrate",
    "integrate_local",
    "derivate",
    "antiderivate",
]


def get_axis(self, axis, is_real):
    """Computes the vector 'axis' in the unit required, using conversions and symmetries if needed.
    Parameters
    ----------
    self: RequestedAxis
        a RequestedAxis object
    axis: Axis
        an Axis object
    """

    if self.operation is not None:
        module = import_module("SciDataTool.Functions.conversions")
        func = getattr(module, self.operation)  # Conversion function
    if isinstance(axis, DataPattern):
        self.is_pattern = True
        self.rebuild_indices = axis.rebuild_indices
        self.is_step = axis.is_step
    is_components = getattr(axis, "is_components", False)
    if (
        is_components and axis.is_overlay and self.extension == "whole"
    ):  # To overlay curves in plots
        self.extension = "list"
        self.indices = [":"]
    if is_components:
        values = axis.get_values(unit=self.unit)
        if self.indices is not None:
            if self.indices[0] == ":":
                self.values = values
                self.indices = list(range(len(values)))
            elif isinstance(self.indices[0], str):  # N largest
                self.extension = "list"
                N = int(self.indices[0])
                # Sort values
                if hasattr(axis, "sort_indices") and axis.sort_indices is not None:
                    self.indices = axis.sort_indices[:N]
                    self.values = values[self.indices]
                else:
                    self.indices = argsort(values)[::-1][:N].tolist()
                    self.values = values[self.indices]
            else:
                self.values = values[self.indices]
                self.extension = "list"
        elif self.extension not in operation_list and self.extension != "list":
            self.values = values
            self.extension = "whole"
        else:
            self.values = values

    else:
        if self.extension == "pattern":
            if not self.is_pattern:
                raise AxisError("[pattern] cannot be called with non DataPattern axis")
            else:
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
                self.extension = "smallestperiod"
        elif self.extension == "smallestperiod":
            # if isinstance(axis, DataPattern):
            #     raise AxisError(
            #         "[smallestperiod] cannot be called with DataPattern axis"
            #     )
            # else:
            is_smallestperiod = True
            is_oneperiod = False
            is_antiperiod = False
        elif self.extension == "antiperiod":
            if isinstance(axis, DataPattern):
                raise AxisError("[antiperiod] cannot be called with DataPattern axis")
            else:
                is_smallestperiod = False
                is_oneperiod = False
                is_antiperiod = True
        elif self.extension == "oneperiod" or self.transform == "fft":
            if isinstance(axis, DataPattern):
                raise AxisError("[oneperiod] cannot be called with DataPattern axis")
            else:
                is_smallestperiod = False
                is_oneperiod = True
                is_antiperiod = False
        elif self.extension in operation_list:
            # Remove periodicities in case of DataPattern otherwise operations can be applied on periodic signals
            is_smallestperiod = not self.is_pattern
            is_oneperiod = False
            is_antiperiod = False
        # Ignore symmetries if fft axis
        elif self.name == "freqs" or self.name == "wavenumber":
            is_smallestperiod = True
            is_oneperiod = False
            is_antiperiod = False
        else:
            if self.input_data is not None and not self.is_step:
                # Check if symmetries need to be reconstructed to match input_data
                axis_values = axis.get_values(
                    is_smallestperiod=True,
                    operation=self.operation,
                    is_real=is_real,
                )
                if min(self.input_data) >= min(axis_values) and max(
                    self.input_data
                ) <= max(axis_values):
                    is_smallestperiod = True
                    is_oneperiod = False
                    is_antiperiod = False
                else:
                    axis_values = axis.get_values(
                        is_oneperiod=True,
                        operation=self.operation,
                        is_real=is_real,
                    )
                    if min(self.input_data) >= min(axis_values) and max(
                        self.input_data
                    ) <= max(axis_values):
                        is_smallestperiod = False
                        is_oneperiod = True
                        is_antiperiod = False
                        self.extension = "oneperiod"
                    else:
                        is_smallestperiod = False
                        is_oneperiod = False
                        is_antiperiod = False
                        if not self.is_pattern:
                            self.extension = "interval"
            elif self.transform == "ifft":  # Ignore symmetries in ifft case
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
            else:
                is_smallestperiod = False
                is_oneperiod = False
                is_antiperiod = False
        # Get original values of the axis including unit and normalizations
        # Store before normalization/operation
        if (
            self.unit != self.corr_unit
            and self.unit != "SI"
            or self.operation is not None
        ):
            self.corr_values = array(
                axis.get_values(
                    is_oneperiod=is_oneperiod,
                    is_antiperiod=is_antiperiod,
                    is_smallestperiod=is_smallestperiod,
                )
            )
        # Prepare unit
        # if self.corr_unit is None:
        #     self.corr_unit = self.unit
        values = axis.get_values(
            is_oneperiod=is_oneperiod,
            is_antiperiod=is_antiperiod,
            is_smallestperiod=is_smallestperiod,
            unit=self.unit,
            operation=self.operation,
            is_real=is_real,
            corr_unit=self.corr_unit,
        )

        # Rebuild symmetries in fft case
        if self.transform == "fft":
            if "period" in axis.symmetries:
                if axis.name != "time":
                    values = values * axis.symmetries["period"]
            elif "antiperiod" in axis.symmetries:
                if axis.name != "time":
                    values = values * axis.symmetries["antiperiod"] / 2
        # Rebuild symmetries in ifft case
        if self.transform == "ifft":
            if self.extension not in [
                "smallestperiod",
                "oneperiod",
                "antiperiod",
                "max",
                "min",
                "sum",
                "mean",
                "rms",
                "rss",
                "integrate",
                "integrate_local",
                "derivate",
                "antiderivate",
                "axis_data",
            ]:
                values = rebuild_symmetries_axis(values, axis.symmetries)
        # Interpolate axis with input data
        if self.input_data is None:
            self.values = values
        else:
            if len(self.input_data) == 2 and self.extension != "axis_data":
                indices = [
                    i
                    for i, x in enumerate(values)
                    if x >= self.input_data[0] and x <= self.input_data[-1]
                ]
                if self.indices is None:
                    self.indices = indices
                else:
                    indices_new = []
                    for i in self.indices:
                        if i in indices:
                            indices_new.append(i)
                    self.indices = indices_new
                self.input_data = None
            else:
                self.values = values
        if self.indices is not None:
            if isinstance(self.indices[0], str):  # N largest
                N = int(self.indices[0])
                # Sort values
                if hasattr(axis, "sort_indices") and axis.sort_indices is not None:
                    self.indices = axis.sort_indices[:N]
                else:
                    self.indices = argsort(values)[::-1][:N].tolist()
            self.values = values[self.indices]
            if self.extension in operation_list:
                self.indices = None



def to_xyz(self):
    """Performs the corrdinate change and stores the resulting field in a VectorField object.
    Parameters
    ----------
    self : VectorField
        a VectorField object
    Returns
    -------
    a VectorField object
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.VectorField", fromlist=["VectorField"])
    VectorField = getattr(module, "VectorField")

    if "comp_x" in self.components or "comp_y" in self.components:
        return self.copy()

    else:
        # Coordinate transform
        arg_list = [
            axis.name
            if axis.name in ["freqs", "wavenumber"]
            else axis.name + "[smallestperiod]"
            for axis in self.components["radial"].axes
        ]
        result = self.get_xyz_along(*arg_list, is_squeeze=False)
        # Store in new VectorField
        comp_dict = dict()

        Comp_x = self.components["radial"].copy()
        Comp_x.name = (
            self.components["radial"].name.lower().replace("radial ", "")
            + " along x-axis"
        )
        Comp_x.symbol = self.components["radial"].symbol.replace("_r", "_x")
        Comp_x.values = result["comp_x"]
        comp_dict["comp_x"] = Comp_x

        Comp_y = self.components["radial"].copy()
        Comp_y.name = (
            self.components["radial"].name.lower().replace("radial ", "")
            + " along y-axis"
        )
        Comp_y.symbol = self.components["radial"].symbol.replace("_r", "_y")
        Comp_y.values = result["comp_y"]
        comp_dict["comp_y"] = Comp_y

        if "axial" in self.components:
            Comp_z = self.components["axial"].copy()
            comp_dict["comp_z"] = Comp_z

        return VectorField(name=self.name, symbol=self.symbol, components=comp_dict)



import numpy as np

from SciDataTool.Functions.change_referential import (
    change_referential_spectrum,
    change_referential_waveform,
)
from SciDataTool.Functions.Load.import_class import import_class


def change_referential(
    self,
    rotation_speed,
    is_waveform,
    atol=1e-9,
    axes_list_change=list(),
    arg_list=list(),
    freqs_new=np.array([]),
    I1=np.array([]),
    Irf_un=np.array([]),
    sym_t_new=dict(),
    ta_in=tuple(),
    ta_out=tuple(),
    atol_freq=1e-6,
):
    """Change referential of a SciDataTool DataND object

    Parameters
    ----------
    self : DataND
        a DataND object
    rotation_speed : float
        rotation speed [rpm]
    is_waveform: bool
        True to perform referential change in waveform domain (time/space) otherwise perform it in spectral domain (freqs/wavenumber)
    atol: float
        Absolute tolerance under which amplitudes are assumed to be 0
    axes_list_change : [Data]
        Axes list to be stored in referential changed DataFreq
    arg_list: list
        List of strings to call get_along
    freqs_new : ndarray
        frequencies vector in the new rotating referential
    I1 : ndarray
        Array of component indices in new spectrum
    Irf_un: ndarray
        Array of indices of unique frequency/wavenumber couples
    ta_in: tuple
        Tuple of input time/angle meshgrids
    ta_out: tuple
        Tuple of output time/angle meshgrids
    atol_freq: float
        Absolute tolerance under which frequencies are assumed to be equal

    Returns
    -------
    data_new : DataND
        a DataND object in new referential
    axes_list_change : [Data]
        Axes list to be stored in changed DataND
    arg_list: list
        List of strings to call get_along
    freqs_new : ndarray
        frequencies vector in the new rotating referential
    I1 : ndarray
        Array of component indices in new spectrum
    Irf_un: ndarray
        Array of indices of unique frequency/wavenumber couples
    ta_in: tuple
        Tuple of input time/angle meshgrids
    ta_out: tuple
        Tuple of output time/angle meshgrids

    """

    Data1D = import_class("SciDataTool.Classes", "Data1D")
    DataLinspace = import_class("SciDataTool.Classes", "DataLinspace")
    DataPattern = import_class("SciDataTool.Classes", "DataPattern")

    if is_waveform:
        DataClass = import_class("SciDataTool.Classes", "DataTime")
    else:
        DataClass = import_class("SciDataTool.Classes", "DataFreq")

    # Get axes list
    axes_list = self.get_axes()
    axes_name = [axis.name for axis in axes_list]

    if len(arg_list) == 0:
        # Get fft along frequency axis (keep other axes)
        arg_list = list()
        for axis in axes_list:
            if axis.name == "time":
                if is_waveform:
                    arg_list.append("time[oneperiod]")
                else:
                    arg_list.append("freqs")
            elif axis.name == "angle":
                if is_waveform:
                    arg_list.append("angle[oneperiod]")
                else:
                    arg_list.append("wavenumber")
            elif axis.name == "z" and isinstance(axis, DataPattern):
                arg_list.append("z[smallestpattern]")
            else:
                arg_list.append(axis.name)

    # Referential changed is requested in waveform domain if time is in arg_list
    is_waveform = "time[oneperiod]" in arg_list

    if not is_waveform and hasattr(self, "is_real"):
        # Force non real data to get all spectrum
        is_real = self.is_real
        self.is_real = False
    else:
        is_real = True

    # Extract data from vectorfield on one period
    result = self.get_along(*arg_list, is_squeeze=False)
    values = result[self.symbol]

    if is_waveform:
        # Get axes values
        time0 = result["time"]
        angle0 = result["angle"]

        # Check requested time anti-periodicity
        is_aper_t = "antiperiod" in sym_t_new

        # Get angle periodicity on one period
        per_a, is_aper_a = axes_list[1].get_periodicity()
        per_a = int(per_a / 2) if is_aper_a else per_a

        val_new, time_new, angle_new, ta_in, ta_out = change_referential_waveform(
            values,
            time0,
            angle0,
            rotation_speed,
            is_aper_a,
            is_aper_t,
            ta_in,
            ta_out,
        )
    else:
        freqs = result["freqs"]
        wavenumber = result["wavenumber"]

        val_new, freqs_new, I1, Irf_un = change_referential_spectrum(
            freqs,
            wavenumber,
            rotation_speed,
            values,
            atol=atol,
            freqs_new=freqs_new,
            I1=I1,
            Irf_un=Irf_un,
            is_double_f0=is_real and "freqs" in axes_name,
            atol_freq=atol_freq,
        )

        if is_real:
            # Adapt spectrum amplitude
            if "time" in axes_name:
                # Input data is DataTime, FFT has just been performed in four quadrants
                # so all amplitudes are multiplied by 2 after restriction to positive frequencies
                coeff = 2
            else:
                # Input data is DataFreq, since data is a real spectrum no need to multiply again
                # all harmonics by two
                coeff = 1

            # Restrict spectrum to positive frequencies only and multiply amplitudes by coeff
            Ifreq_pos = freqs_new >= 0
            freqs_store = freqs_new[Ifreq_pos]
            val_new = val_new[Ifreq_pos] * coeff
            val_new[0] = val_new[0] / coeff
        else:
            freqs_store = freqs_new

    if len(axes_list_change) == 0:
        axes_list_change = list()
        for axis in axes_list:

            if axis.name == "time" and is_waveform:
                axis_new = axis.copy()
                axis_new.symmetries = sym_t_new
                axis_new.number = time_new.size
                if isinstance(axis_new, DataLinspace):
                    axis_new.final = time_new[-1]
                    axis_new.include_endpoint = True
                else:
                    axis_new.values = time0

            elif axis.name == "angle" and is_waveform:
                axis_new = axis.copy()
                if is_aper_a:
                    axis_new.symmetries = {"antiperiod": int(2 * per_a)}
                else:
                    axis_new.symmetries = {"period": per_a}
                axis_new.number = angle_new.size
                if isinstance(axis_new, DataLinspace):
                    axis_new.final = angle_new[-1]
                    axis_new.include_endpoint = True
                else:
                    axis_new.values = angle_new

            elif axis.name == "freqs" or (axis.name == "time" and not is_waveform):
                axis_new = Data1D(
                    name="freqs",
                    unit="Hz",
                    values=freqs_store,
                    normalizations=axis.normalizations.copy(),
                    symmetries=dict(),  # Don't keep time symmetries, FFT inverse requests axis_data
                )

            elif axis.name == "wavenumber" or (
                axis.name == "angle" and not is_waveform
            ):
                axis_new = Data1D(
                    name="wavenumber",
                    unit="dimless",
                    values=wavenumber,
                    normalizations=axis.normalizations.copy(),
                    symmetries=axis.symmetries.copy(),
                )

            else:
                axis_new = axis.copy()

            axes_list_change.append(axis_new)

    data_new = DataClass(
        name=self.name,
        symbol=self.symbol,
        unit=self.unit,
        axes=axes_list_change,
        values=val_new,
        normalizations=self.normalizations.copy(),
        is_real=is_real,
    )

    if not is_waveform and hasattr(self, "is_real"):
        # Reset data
        self.is_real = is_real

    return data_new, axes_list_change, arg_list, freqs_new, I1, Irf_un, ta_in, ta_out



import matplotlib.pyplot as plt
from numpy import arange, nanmax, nanmin, frombuffer
import imageio
from ...GUI.DDataPlotter.DDataPlotter import PARAM_3D
from SciDataTool.Functions.Plot import fft_dict, ifft_dict


def plot_2D_Data_Animated(
    self, animated_axis, suptitle_ref, *param_list, nb_frames=50, fps=10, **param_dict
):
    """Gen

    Parameters
    ----------
    animated_axis : str
        The field will be animated along this axis
    nb_frames : int
        number of frames used to build the gif
    fps: int
        frames displayed per second
    """
    # Relative import of DataPattern to prevent circular import
    module = __import__("SciDataTool.Classes.DataPattern", fromlist=["DataPattern"])
    DataPattern = getattr(module, "DataPattern")

    # Making sure that we have the right argument for a plot2D
    plot_options = param_dict.copy()
    for param in PARAM_3D:
        if param in plot_options:
            del plot_options[param]

    # Detecting if animated axis is a DataPattern, if true changing the input given to the function
    for ax_obj in self.get_axes():
        if (
            ax_obj.name == animated_axis.split("[")[0]
            or animated_axis.split("[")[0] in fft_dict
            and fft_dict[animated_axis.split("[")[0]] == ax_obj.name
            or animated_axis.split("[")[0] in ifft_dict
            and ifft_dict[animated_axis.split("[")[0]] == ax_obj.name
        ):
            animated_axis_obj = ax_obj
            break

    if isinstance(animated_axis_obj, DataPattern):
        # Removing "[one_period]" as it is not available with a DataPattern
        animated_axis_unit = "{" + animated_axis.split("{")[1]
        animated_axis = animated_axis.split("[")[0] + animated_axis_unit

    if "freqs" in param_list or "wavenumber" in param_list:
        result = self.get_magnitude_along(
            animated_axis, *param_list, unit=param_dict["unit"]
        )
    else:
        result = self.get_along(animated_axis, *param_list, unit=param_dict["unit"])

    animated_axis_unit = "{" + animated_axis.split("{")[1]
    animated_axis = animated_axis.split("{")[0].split("[")[0]

    # Creating a list of frames that will need to create the animation
    if isinstance(animated_axis_obj, DataPattern):
        frames_list = animated_axis_obj.get_values()
        frames_list = [
            "[" + str(idx_frame) + "]" for idx_frame in range(len(frames_list))
        ]
        fps = 1
    else:
        value_max = nanmax(result[animated_axis])
        value_min = nanmin(result[animated_axis])
        variation_step = (value_max - value_min) / nb_frames

        frames_list = arange(start=value_min, stop=value_max, step=variation_step)
        frames_list = ["=" + str(frame) for frame in frames_list]

    # detecting if we are animating a regular plot (=> computing limit for y) or a "fft" plot (=> limit already set)
    if plot_options["y_min"] == None or plot_options["y_max"] == None:
        # Setting the options of the plot
        y_max = nanmax(result[self.symbol])
        y_min = nanmin(result[self.symbol])
        marge = (
            y_max - y_min
        ) * 0.05  # 5% of the height of plot to add to the border top/bottom of gif
        plot_options["y_min"] = nanmin(result[self.symbol]) - abs(marge)
        plot_options["y_max"] = nanmax(result[self.symbol]) + abs(marge)

    plot_options["is_show_fig"] = False

    # Getting the name of the gif
    save_path = plot_options["save_path"].replace(".png", ".gif")
    plot_options["save_path"] = None

    images = list()  # List of images used to build the gif
    for val in frames_list:
        # plotting image
        self.plot_2D_Data(
            *param_list, animated_axis + val + animated_axis_unit, **plot_options
        )
        # Getting the figure generated with plot_2D_DATA
        fig = plt.gcf()

        # Adding the suptitle of the figure if there is one
        if suptitle_ref != "":
            fig.suptitle(suptitle_ref)

        fig.canvas.draw()
        image = frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
        image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        images.append(image)

        # Creating the gif
        plt.close(fig)

    imageio.mimsave(save_path, images, format="GIF-PIL", fps=fps)



from PySide2.QtWidgets import QWidget

from SciDataTool.GUI.WAxisSelector.Ui_WAxisSelector import Ui_WAxisSelector
from SciDataTool.GUI.WFilter.WFilter import WFilter
from PySide2.QtCore import Signal
from SciDataTool.Functions.Plot import (
    unit_dict,
    norm_name_dict,
    axis_norm_dict,
    axes_dict,
    fft_dict,
    ifft_dict,
)
from SciDataTool.GUI import update_cb_enable
from SciDataTool.Functions.Load.import_class import import_class


class WAxisSelector(Ui_WAxisSelector, QWidget):
    """Widget to select the axis to plot"""

    refreshNeeded = Signal()
    axisChanged = Signal()
    actionChanged = Signal()

    def __init__(self, parent=None, path_to_image=None):
        """Initialize the arguments, linking the buttons and setting up the UI

        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        parent : QWidget
            The parent widget
        """

        # Build the interface according to the .ui file
        QWidget.__init__(self, parent)
        self.setupUi(self)

        self.name = "X"  # Name of the axis
        self.axes_list = list()  # List of the different axes of the DataND object
        self.norm_list = None  # List of available normalizations for each axis
        self.indices = None
        self.path_to_image = path_to_image

        self.axis_selected = "None"  # Name of the axis selected (time, angle...)
        self.norm = None  # Name of the unit of the axis (s,m...)
        self.b_filter.setEnabled(False)
        self.current_dialog = None

        self.c_axis.currentTextChanged.connect(self.update_axis)
        self.c_action.currentTextChanged.connect(self.update_action)
        self.c_unit.currentTextChanged.connect(self.update_unit)
        self.b_filter.clicked.connect(self.open_filter)

    def get_axes_name(self):
        """Method that return the axes that can be selected
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        Output
        ---------
        list
            name of the axes avalaible
        """
        return self.axes_list

    def get_axis_unit_selected(self):
        """Method that return the axis and the unit currently selected so that we can use it in the plot method
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        Output
        ---------
        string
            name of the current axis and unit selected in the right format
        """

        axis_unit_selected = self.get_axis_selected()

        # Add indices if necessary
        if self.get_current_action_name() == "Filter":
            if self.indices is not None:
                axis_unit_selected += str(self.indices)
            else:
                axis_unit_selected += "[]"

        if self.norm is not None:  # Add normalization
            axis_unit_selected += "{" + self.norm + "}"

        elif axis_unit_selected != "None":  # adding unit
            axis_unit_selected += "{" + self.c_unit.currentText() + "}"

        return axis_unit_selected

    def get_current_action_name(self):
        """Method that return the action currently selected
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        Output
        ---------
        string
            name of the current action selected
        """
        return self.c_action.currentText()

    def get_axis_selected(self):
        """Method that return the name of the axis selected of the WAxisSelector
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        Output
        ---------
        string
            name of the axis selected
        """
        if self.c_axis.currentText() == "None":
            return "None"
        else:
            return self.axis_selected

    def open_filter(self):
        """Open the Filter widget"""
        # Close previous dialog
        if self.current_dialog is not None:
            self.current_dialog.close()
            self.current_dialog.setParent(None)
            self.current_dialog = None

        axis_selected_obj = [
            ax for ax in self.axes_list_obj if ax.name == self.axis_selected
        ][0]

        self.current_dialog = WFilter(
            axis_selected_obj,
            self.indices,
            path_to_image=self.path_to_image,
        )
        self.current_dialog.refreshNeeded.connect(self.update_indices)
        self.current_dialog.show()

    def update_indices(self):
        self.indices = self.current_dialog.indices
        self.refreshNeeded.emit()

    def remove_axis(self, axis_to_remove):
        """Method that remove a given axis from the axis ComboBox.
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        axis_to_remove : string
            name of the axis to remove from c_axis

        """
        # if the axis to remove is wavenumber of freqs then we have to remove angle or time
        if axis_to_remove in ifft_dict:
            axis_to_remove = ifft_dict[axis_to_remove]

        if axis_to_remove in self.axes_list:
            axes_list = self.axes_list.copy()  # Getting the axes available
            axes_list.remove(axis_to_remove)  # Removing the axis selected

            # Building the new ComboBox
            self.c_axis.blockSignals(True)
            index = 0
            if self.c_axis.currentText() != "None":
                index = self.c_axis.currentIndex()
            self.c_axis.clear()

            for ax in axes_list:
                if ax in axes_dict:
                    self.c_axis.addItem(axes_dict[ax])
                else:
                    self.c_axis.addItem(ax)

            update_cb_enable(self.c_axis)
            self.c_axis.setCurrentIndex(index)
            self.c_axis.blockSignals(False)

            self.update_axis(is_refresh=True)

    def set_axis(self, axis):
        """Method that will set the comboboxes to have the axis given as an input when calling the plot method (auto-plot).
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        axis : RequestedAxis
            axis that we want to have in the WAxisSelector
        """
        self.blockSignals(True)
        self.indices = axis.indices
        # Step 1 : Getting the name of the axis and selecting the right combobox (axis and action)
        axis_name = axis.name

        # If the axis is freqs or wavenumber, then we have to select time/angle and fft
        if axis_name in ifft_dict:
            # Selecting the right axis
            self.c_axis.setCurrentIndex(self.c_axis.findText(ifft_dict[axis_name]))
            # Making sure that we select FFT
            self.c_action.setCurrentIndex(1)
        else:
            # Selecting the right axis
            if axis_name in axes_dict:
                self.c_axis.setCurrentIndex(self.c_axis.findText(axes_dict[axis_name]))
            else:
                self.c_axis.setCurrentIndex(self.c_axis.findText(axis_name))

        # Step 2 : Recovering the unit and setting the combobox according to it
        unit_name = axis.unit
        self.c_unit.blockSignals(True)
        self.set_unit()
        if self.c_unit.findText(unit_name) != -1:
            self.c_unit.setCurrentIndex(self.c_unit.findText(unit_name))
        elif unit_name in unit_dict:
            self.c_unit.setCurrentIndex(self.c_unit.findText(unit_dict[unit_name]))
        elif unit_name in norm_name_dict:
            self.c_unit.setCurrentIndex(self.c_unit.findText(norm_name_dict[unit_name]))
        elif unit_name == "SI":
            self.c_unit.setCurrentIndex(0)
        else:
            self.c_unit.setCurrentIndex(self.c_unit.findText(unit_name))
        self.c_unit.blockSignals(False)

        self.blockSignals(False)

    def set_axis_options(self, axes_list):
        """Method that will put the axes of data in the combobox of the widget
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        axes_list : list
            A list of axis object from the Data object that we want to plot

        """
        # Dynamic import to avoid import loop
        DataPattern = import_class("SciDataTool.Classes", "DataPattern")

        self.c_axis.blockSignals(True)
        # Step 1 : Getting the name of the different axes of the DataND object
        self.axes_list = [
            axis.name
            for axis in axes_list
            if axis.is_overlay == False
            and not (isinstance(axis, DataPattern) and len(axis.unique_indices) == 1)
        ]  # Remove overlay axes + slice axes with single slice
        self.norm_list = [
            list(axis.normalizations.keys())
            for axis in axes_list
            if axis.is_overlay == False
            and not (isinstance(axis, DataPattern) and len(axis.unique_indices) == 1)
        ]

        # Adding a safety, so that we cannot have frequency or wavenumber inside axes_list (we should have time and angle instead)
        for i in range(len(self.axes_list)):
            if self.axes_list[i] in ifft_dict:
                self.axes_list[i] = ifft_dict[self.axes_list[i]]

        # At least one axis must be selected => impossible to have none for X axis
        if self.name.lower() != "x":
            self.axes_list.insert(0, "None")
            self.norm_list.insert(0, None)

        # Step 2 : Replacing the items inside of the ComboBox with the axes recovered
        self.c_axis.clear()
        for ax in self.axes_list:
            if ax in ifft_dict:
                self.c_axis.addItem(ifft_dict[ax])
            else:
                if ax in axes_dict:
                    self.c_axis.addItem(axes_dict[ax])
                else:
                    self.c_axis.addItem(ax)

        update_cb_enable(self.c_axis)
        self.c_axis.blockSignals(False)

        # Step 3 : Modifying axis_selected
        if self.c_axis.currentText() in [axes_dict[key] for key in axes_dict]:
            self.axis_selected = [key for key in axes_dict][
                [axes_dict[key] for key in axes_dict].index(self.c_axis.currentText())
            ]
        else:
            self.axis_selected = self.c_axis.currentText()

        self.c_axis.view().setMinimumWidth(max([len(ax) for ax in self.axes_list]) * 6)

        if self.c_axis.currentText() == "None":
            self.c_action.setDisabled(True)
        else:
            self.c_action.setDisabled(False)

    def set_name(self, axis_name):
        """Method to change of the label of the widget
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        axis_name : string
            string that we will use to set the in_name of the widget

        """

        self.name = axis_name
        self.in_name.setText(axis_name)

    def set_action(self, action):
        """Method that set the action of the WAxisSelector
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        action : string
            name of the new action"""

        action_list = [self.c_action.itemText(i) for i in range(self.c_action.count())]
        self.c_action.blockSignals(True)
        if action in action_list and action != "Filter":
            self.c_action.setCurrentIndex(action_list.index(action))
        self.c_action.blockSignals(False)

    def set_unit(self, is_refresh=True):
        """Method that update the unit comboxbox according to the axis selected in the other combobox.
           We can also give the axis selected and put its units inside the combobox
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        """
        self.c_unit.blockSignals(True)
        # Adding the right units according to a dictionnary
        if self.axis_selected == "None":
            # If the axis is not selected, then we can not choose the unit
            self.c_unit.clear()
            self.c_unit.setDisabled(True)
        else:
            unit_index = self.c_unit.currentIndex()
            self.c_unit.setDisabled(False)
            self.c_unit.clear()

            # Adding the axis unit + available normalizations
            if self.axis_selected in unit_dict:
                self.c_unit.addItem(unit_dict[self.axis_selected])
            norms = self.norm_list[self.axes_list.index(self.axis_selected)]
            norm_id = None
            if norms is not None:
                for norm in norms:
                    if norm in axis_norm_dict[self.axis_selected]:
                        self.c_unit.addItem(norm_name_dict[norm])  # Add longer names
                        if "_id" in norm:
                            norm_id = norm_name_dict[norm]

            # Use tooth index as only normalization
            if norm_id is not None:
                self.c_unit.clear()
                self.c_unit.addItem(norm_id)

            # Modifying the size of the list according to the units available (adapting it when normalization)
            cb_width = 0
            for idx_unit in range(self.c_unit.count()):
                self.c_unit.setCurrentIndex(idx_unit)
                if len(self.c_unit.currentText()) > cb_width:
                    cb_width = len(self.c_unit.currentText())
            if unit_index != -1:
                self.c_unit.setCurrentIndex(unit_index)

            self.c_unit.view().setMinimumWidth(cb_width * 8)

        update_cb_enable(self.c_unit)
        self.c_unit.blockSignals(False)
        self.update_unit(is_refresh=is_refresh)

    def update(self, axes_list, axis_name="X"):
        """Method used to update the widget by calling the other method for the label, the axes and the units
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        axes_list : list
            A list of axis object from the Data object that we want to plot
        axis_name : string
            string that will set the text of in_name (=name of the axis)
        """
        self.axes_list_obj = axes_list
        self.set_name(axis_name)
        self.set_axis_options(axes_list)
        self.update_axis()
        self.set_unit()

    def update_axis(self, text=None, is_refresh=True):
        """Method called when an axis is changed that change axis_selected, the units available and the action combobox.
        It will also emit a signal used in WAxisManager.
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object
        """
        self.c_action.blockSignals(True)
        if is_refresh:
            self.c_action.setCurrentIndex(0)

        if self.c_axis.currentText() == "None":
            self.c_action.setDisabled(True)
        else:
            self.c_action.setDisabled(False)

        # Updating the units and the axis selected
        # Making sure that self.axis_selected is a "tag" and not a "label". Example : z instead of axial direction
        if self.c_axis.currentText() in [axes_dict[key] for key in axes_dict]:
            self.axis_selected = [key for key in axes_dict][
                [axes_dict[key] for key in axes_dict].index(self.c_axis.currentText())
            ]
        else:
            self.axis_selected = self.c_axis.currentText()

        if not is_refresh:
            self.set_unit()

        # Updating the action combobox
        # Handling specific case to disable certain parts of the GUI
        if self.c_axis.currentText() == "None":
            self.c_action.setDisabled(True)
        else:
            self.c_action.setDisabled(False)

        if self.axis_selected in fft_dict:
            action = ["None", "FFT"]
        elif (
            self.axis_selected in [axis.name for axis in self.axes_list_obj]
            and self.axes_list_obj[
                [axis.name for axis in self.axes_list_obj].index(self.axis_selected)
            ].is_components
        ):
            action = ["None", "Filter"]
        else:
            action = ["None"]
        self.c_action.clear()
        self.c_action.addItems(action)
        update_cb_enable(self.c_action)
        self.c_action.blockSignals(False)

        self.c_action.view().setMinimumWidth(max([len(ac) for ac in action]) * 6)

        if "Filter" in action:
            self.b_filter.show()
        else:
            sp_retain = self.b_filter.sizePolicy()
            sp_retain.setRetainSizeWhenHidden(True)
            self.b_filter.setSizePolicy(sp_retain)
            self.b_filter.hide()

        # Emitting the signals
        if is_refresh:
            self.axisChanged.emit()

    def update_action(self):
        """Method called when an action is changed that will change axis_selected,
        update the units available and emit a signal.
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object

        """
        # If the action selected is filter, then we enable the button
        if self.c_action.currentText() == "Filter":
            self.b_filter.setEnabled(True)

        else:
            self.b_filter.setEnabled(False)

        # Converting the axes according to action selected if possible/necessary
        if self.c_action.currentText() == "FFT" and self.axis_selected in fft_dict:
            self.axes_list.insert(
                self.axes_list.index(self.axis_selected), fft_dict[self.axis_selected]
            )
            self.axes_list.remove(self.axis_selected)
            self.axis_selected = fft_dict[self.axis_selected]

        elif self.c_action.currentText() == "None" and self.axis_selected in ifft_dict:
            self.axes_list.insert(
                self.axes_list.index(self.axis_selected), ifft_dict[self.axis_selected]
            )
            self.axes_list.remove(self.axis_selected)
            self.axis_selected = ifft_dict[self.axis_selected]

        # Handling the case where axis_selected is updated but axes_list is not
        # We check if fft(axis_selected) is in the list and when it is the case we replace it by axis_selected
        if not self.axis_selected in self.axes_list:
            if fft_dict[self.axis_selected] in self.axes_list:
                self.axes_list.insert(
                    self.axes_list.index(fft_dict[self.axis_selected]),
                    self.axis_selected,
                )
                self.axes_list.remove(fft_dict[self.axis_selected])

        # Now that the quantiy has been updated according to the action, we can set the units and emit the signals
        self.set_unit(is_refresh=False)

        self.actionChanged.emit()

    def update_unit(self, is_refresh=True):
        """Method called when a new unit is selected so that we can update self.norm
        Parameters
        ----------
        self : WAxisSelector
            a WAxisSelector object

        """
        normalization = self.c_unit.currentText()
        is_match = False
        # Replace normalization title with normalization name
        for norm in norm_name_dict:
            if normalization == norm_name_dict[norm]:
                self.norm = norm
                is_match = True
        if not is_match:
            self.norm = None
        if is_refresh:
            self.refreshNeeded.emit()



from numpy import any as np_any

from SciDataTool.Functions.sum_convolution import get_sum_indices, get_sum_amplitudes


def sum(
    self,
    other,
    tol_freq=1e-4,
    name=None,
    symbol=None,
    unit=None,
    normalizations=None,
):
    """Merge two DataFreq objects

    Parameters
    ----------
    self : DataFreq
        A DataFreq object
    other : DataFreq
        Another DataFreq object
    tol_freq: float
        Absolute tolerance value to filter harmonic orders by their frequency value
    name: str
        name
    symbol: str
        symbol
    unit: str
        unit
    normalizations: {Normalization}
        Dict of normalization objects

    Returns
    -------
    result : Spectrum
        DataFreq object resulting from summing both DataFreq objects
    """

    if not isinstance(other, type(self)):
        raise Exception("other is not a DataFreq object")

    if len(self.axes) > 1:
        raise Exception("self contains more than one axis")

    if len(other.axes) > 1:
        raise Exception("other contains more than one axis")

    if self.axes[0].name != "freqs":
        raise Exception("self axis is not frequency axis")
    else:
        freqs1 = self.axes[0].get_values()
        if np_any(freqs1) < 0:
            raise Exception("self contains negative frequency values")

    if other.axes[0].name != "freqs":
        raise Exception("other axis is not frequency axis")
    else:
        freqs2 = other.axes[0].get_values()
        if np_any(freqs2) < 0:
            raise Exception("other contains negative frequency values")

    # Fill metadata
    if name is None:
        name = self.name
    if symbol is None:
        symbol = self.symbol
    if unit is None:
        unit = self.unit
    if normalizations is None:
        normalizations = self.normalizations

    # Compute spectrum orders resulting from convolution
    freqs_un, I0b = get_sum_indices(freqs1, freqs2, tol_freq)

    # Compute spectrum amplitudes resulting from convolution
    amp = get_sum_amplitudes(self.values, other.values, I0b)

    # Create Frequency axis
    Freqs = self.axes[0].copy()
    Freqs.values = freqs_un

    # Create DataFreq resulting from convolution
    result = type(self)(
        name=name,
        unit=unit,
        symbol=symbol,
        axes=[Freqs],
        values=amp,
        normalizations=normalizations,
    )

    return result



# -*- coding: utf-8 -*-

# File generated according to WFilter.ui
# WARNING! All changes made in this file will be lost!
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from SciDataTool.GUI.Tools.CheckBox import CheckBox


class Ui_WFilter(object):
    def setupUi(self, WFilter):
        if not WFilter.objectName():
            WFilter.setObjectName("WFilter")
        WFilter.resize(831, 644)
        WFilter.setMinimumSize(QSize(630, 470))
        WFilter.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout = QGridLayout(WFilter)
        self.gridLayout.setObjectName("gridLayout")
        self.tab_indices = QTableView(WFilter)
        self.tab_indices.setObjectName("tab_indices")

        self.gridLayout.addWidget(self.tab_indices, 0, 0, 1, 1)

        self.cb_all = CheckBox(WFilter)
        self.cb_all.setObjectName("cb_all")
        self.cb_all.setTristate(True)

        self.gridLayout.addWidget(self.cb_all, 1, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.b_Ok = QPushButton(WFilter)
        self.b_Ok.setObjectName("b_Ok")

        self.horizontalLayout.addWidget(self.b_Ok)

        self.b_cancel = QPushButton(WFilter)
        self.b_cancel.setObjectName("b_cancel")

        self.horizontalLayout.addWidget(self.b_cancel)

        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 1)

        self.retranslateUi(WFilter)

        QMetaObject.connectSlotsByName(WFilter)

    # setupUi

    def retranslateUi(self, WFilter):
        WFilter.setWindowTitle(QCoreApplication.translate("WFilter", "WFilter", None))
        self.cb_all.setText(
            QCoreApplication.translate("WFilter", "Select/Deselect all", None)
        )
        self.b_Ok.setText(QCoreApplication.translate("WFilter", "Ok", None))
        self.b_cancel.setText(QCoreApplication.translate("WFilter", "Cancel", None))

    # retranslateUi



import numpy as np


def change_referential(
    self, rotation_speed, is_waveform, atol=1e-9, sym_t_new=dict(), atol_freq=1e-6
):
    """Change referential of vectorfield given input rotation speed (algebric to include rotation direction)

    Parameters
    ----------
    self : VectorField
        A VectorField object
    rotation_speed : float
        rotation speed [rpm]
    is_waveform: bool
        True to perform referential change in waveform domain (time/space) otherwise perform it in spectral domain (freqs/wavenumber)
    atol: float
        Absolute tolerance under which amplitudes are assumed to be 0
    sym_t_new: dict
        Dict of symmetries in new referential
    atol_freq: float
        Absolute tolerance under which frequencies are assumed to be equal

    Returns
    -------
    vf_new : VectorField
        Vectorfield converted in new referential
    """

    # Init referential changed VectorField
    vf_new = type(self)(name=self.name, symbol=self.symbol, components=dict())

    # Loop on components to convert them to new referential
    axes_list_change = list()
    arg_list = list()
    freqs_new = np.array([])
    I1 = np.array([])
    Irf_un = np.array([])
    Imask = np.array([])
    ta_in = tuple()
    ta_out = tuple()
    for comp, data in self.components.items():

        # Change referential
        (
            data_new,
            axes_list_change,
            arg_list,
            freqs_new,
            I1,
            Irf_un,
            ta_in,
            ta_out,
        ) = data.change_referential(
            rotation_speed,
            is_waveform,
            atol=0,
            axes_list_change=axes_list_change,
            arg_list=arg_list,
            freqs_new=freqs_new,
            I1=I1,
            Irf_un=Irf_un,
            sym_t_new=sym_t_new,
            ta_in=ta_in,
            ta_out=ta_out,
            atol_freq=atol_freq,
        )

        # Filter harmonics that are below input absolute tolerance
        if atol > 0 and not is_waveform:

            # Copy axis not to change axis in axes_list_change
            data_new.axes = [axis.copy() for axis in data_new.axes]

            # Filter harmonics that are below input absolute tolerance
            if Imask.size == 0:
                val = np.abs(data_new.values)
                val_max = np.max(val)
                while val.ndim > 1:
                    val = np.sum(val, axis=-1)
                Imask = val > val_max * atol
            data_new.axes[0].values = data_new.axes[0].values[Imask]
            data_new.values = data_new.values[Imask, ...]

        # Store referential changed data in VectorField
        vf_new.components[comp] = data_new

    return vf_new



from PySide2.QtCore import QObject, Signal, Slot
from SciDataTool.GUI.Tools.ThreadKillable import ThreadKillable
import multiprocessing
from SciDataTool.GUI.Tools.GifHandler import GifHandler
from logging.handlers import QueueListener

from pytest import param


def save_gif(
    queue, widget, main_widget, gif, plot_input, data_selection, is_3D, suptitle_ref
):

    animated_axis = plot_input.pop(0)

    param_dict = widget.param_dict.copy()

    param_dict["save_path"] = gif

    [
        _,
        _,
        _,
        output_range,
    ] = widget.get_plot_info()

    if "type_plot" in param_dict and param_dict["type_plot"] == "quiver":
        param_dict = main_widget.plot_arg_dict.copy()
        param_dict["save_path"] = gif
        if widget.w_vect_selector.get_component_selected() != "all":
            component_list = [widget.w_vect_selector.get_component_selected()]
        else:
            component_list = None
        angle_str = "angle"
        if "is_smallestperiod" in param_dict:
            if param_dict["is_smallestperiod"]:
                angle_str = "angle[smallestperiod]"
            del param_dict["is_smallestperiod"]
        if "fig" in param_dict:
            fig = param_dict["fig"]
            ax = fig.get_axes()[0]
            del param_dict["fig"]
        else:
            fig = None
            ax = None
        if "colormap" in param_dict:
            del param_dict["colormap"]
        main_widget.data_orig.plot_2D_Data_Animated(
            animated_axis,
            *[*[angle_str], *data_selection],
            **param_dict,
            unit=output_range["unit"],
            fig=fig,
            ax=ax,
            y_min=output_range["min"],
            y_max=output_range["max"],
            component_list=component_list,
        )

    else:

        param_dict["unit"] = output_range["unit"]

        if "component_list" in param_dict:
            param_dict.pop("component_list")

        if is_3D:
            widget.data.plot_3D_Data_Animated(
                animated_axis, suptitle_ref, *plot_input, **param_dict
            )
        else:
            widget.data.plot_2D_Data_Animated(
                animated_axis, suptitle_ref, *plot_input, **param_dict
            )
    queue.put("gif generated")


class SaveGifWorker(QObject):
    """Worker that saves a gif plot"""

    gif_available = Signal()

    def __init__(
        self,
        widget=None,
        main_widget=None,
        gif="",
        plot_input=list(),
        data_selection=list(),
        is_3D=False,
        suptitle_ref="",
    ):
        super().__init__()
        self.widget = widget
        self.main_widget = main_widget
        self.gif = gif
        self.plot_input = plot_input
        self.data_selection = data_selection
        self.is_3D = is_3D
        self.suptitle_ref = suptitle_ref
        self.queue = multiprocessing.Queue()
        # used to check if the is finished
        self.queue_handler = GifHandler(parent=self)
        self.queue_listener = QueueListener(self.queue, self.queue_handler)
        self.queue_listener.start()

    @Slot()
    def run(self):
        # Setting the thread that will start the run simulation
        self.p = ThreadKillable(
            target=save_gif,
            args=(
                self.queue,
                self.widget,
                self.main_widget,
                self.gif,
                self.plot_input,
                self.data_selection,
                self.is_3D,
                self.suptitle_ref,
            ),
        )
        self.p.daemon = True
        self.p.start()

    def kill_worker(self):
        """
        Kill the worker
        Parameters
        ----------
        self: SaveGifWorker
            the worker generating the gif
        Returns
        -------

        """
        if self.queue_listener._thread is not None:
            # Stopping the queue listener
            self.queue_listener.stop()
            self.queue_listener.queue.close()
            self.queue_handler.close()
        # Closing the process (running simu or saving output
        self.p.kill()
        self.p.join()



# -*- coding: utf-8 -*-

# File generated according to WSliceOperator.ui
# WARNING! All changes made in this file will be lost!
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from SciDataTool.GUI.Tools.FloatEdit import FloatEdit
from SciDataTool.GUI.Tools.ButtonLabel import ButtonLabel

from SciDataTool.GUI.Resources import SDT_rc


class Ui_WSliceOperator(object):
    def setupUi(self, WSliceOperator):
        if not WSliceOperator.objectName():
            WSliceOperator.setObjectName("WSliceOperator")
        WSliceOperator.resize(375, 133)
        self.gridLayout = QGridLayout(WSliceOperator)
        self.gridLayout.setObjectName("gridLayout")
        self.b_action = QPushButton(WSliceOperator)
        self.b_action.setObjectName("b_action")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.b_action.sizePolicy().hasHeightForWidth())
        self.b_action.setSizePolicy(sizePolicy)
        self.b_action.setMinimumSize(QSize(0, 30))

        self.gridLayout.addWidget(self.b_action, 2, 0, 1, 1)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.lf_value = FloatEdit(WSliceOperator)
        self.lf_value.setObjectName("lf_value")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(70)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.lf_value.sizePolicy().hasHeightForWidth())
        self.lf_value.setSizePolicy(sizePolicy1)
        self.lf_value.setMinimumSize(QSize(0, 20))
        self.lf_value.setMaximumSize(QSize(70, 16777215))

        self.horizontalLayout_2.addWidget(self.lf_value)

        self.in_unit = QLabel(WSliceOperator)
        self.in_unit.setObjectName("in_unit")

        self.horizontalLayout_2.addWidget(self.in_unit)

        self.slider = QSlider(WSliceOperator)
        self.slider.setObjectName("slider")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.slider.sizePolicy().hasHeightForWidth())
        self.slider.setSizePolicy(sizePolicy2)
        self.slider.setMinimumSize(QSize(0, 20))
        self.slider.setOrientation(Qt.Horizontal)

        self.horizontalLayout_2.addWidget(self.slider)

        self.b_animate = ButtonLabel(WSliceOperator)
        self.b_animate.setObjectName("b_animate")
        sizePolicy3 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.b_animate.sizePolicy().hasHeightForWidth())
        self.b_animate.setSizePolicy(sizePolicy3)
        self.b_animate.setMinimumSize(QSize(20, 20))
        self.b_animate.setMaximumSize(QSize(20, 20))
        self.b_animate.setFrameShadow(QFrame.Sunken)
        self.b_animate.setScaledContents(True)

        self.horizontalLayout_2.addWidget(self.b_animate)

        self.gridLayout.addLayout(self.horizontalLayout_2, 1, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.in_name = QLabel(WSliceOperator)
        self.in_name.setObjectName("in_name")
        sizePolicy.setHeightForWidth(self.in_name.sizePolicy().hasHeightForWidth())
        self.in_name.setSizePolicy(sizePolicy)
        self.in_name.setMinimumSize(QSize(0, 20))

        self.horizontalLayout.addWidget(self.in_name)

        self.c_operation = QComboBox(WSliceOperator)
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.addItem("")
        self.c_operation.setObjectName("c_operation")
        sizePolicy.setHeightForWidth(self.c_operation.sizePolicy().hasHeightForWidth())
        self.c_operation.setSizePolicy(sizePolicy)
        self.c_operation.setMinimumSize(QSize(0, 20))

        self.horizontalLayout.addWidget(self.c_operation)

        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)

        self.l_loading = QLabel(WSliceOperator)
        self.l_loading.setObjectName("l_loading")
        self.l_loading.setEnabled(True)

        self.gridLayout.addWidget(self.l_loading, 3, 0, 1, 1)

        self.retranslateUi(WSliceOperator)

        QMetaObject.connectSlotsByName(WSliceOperator)

    # setupUi

    def retranslateUi(self, WSliceOperator):
        WSliceOperator.setWindowTitle(
            QCoreApplication.translate("WSliceOperator", "WSliceOperator", None)
        )
        self.b_action.setText(
            QCoreApplication.translate("WSliceOperator", "Overlay", None)
        )
        self.lf_value.setText(
            QCoreApplication.translate("WSliceOperator", "0.314", None)
        )
        self.in_unit.setText(QCoreApplication.translate("WSliceOperator", "[m]", None))
        self.b_animate.setText("")
        self.in_name.setText(
            QCoreApplication.translate("WSliceOperator", "angle", None)
        )
        self.c_operation.setItemText(
            0, QCoreApplication.translate("WSliceOperator", "slice", None)
        )
        self.c_operation.setItemText(
            1, QCoreApplication.translate("WSliceOperator", "slice (fft)", None)
        )
        self.c_operation.setItemText(
            2, QCoreApplication.translate("WSliceOperator", "rms", None)
        )
        self.c_operation.setItemText(
            3, QCoreApplication.translate("WSliceOperator", "rss", None)
        )
        self.c_operation.setItemText(
            4, QCoreApplication.translate("WSliceOperator", "sum", None)
        )
        self.c_operation.setItemText(
            5, QCoreApplication.translate("WSliceOperator", "mean", None)
        )
        self.c_operation.setItemText(
            6, QCoreApplication.translate("WSliceOperator", "integrate", None)
        )
        self.c_operation.setItemText(
            7, QCoreApplication.translate("WSliceOperator", "overlay/filter", None)
        )

        self.l_loading.setText(
            QCoreApplication.translate("WSliceOperator", "Generating...", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-

# File generated according to WAxisSelector.ui
# WARNING! All changes made in this file will be lost!
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_WAxisSelector(object):
    def setupUi(self, WAxisSelector):
        if not WAxisSelector.objectName():
            WAxisSelector.setObjectName("WAxisSelector")
        WAxisSelector.resize(152, 160)
        self.verticalLayout = QVBoxLayout(WAxisSelector)
        self.verticalLayout.setObjectName("verticalLayout")
        self.in_name = QLabel(WAxisSelector)
        self.in_name.setObjectName("in_name")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.in_name.sizePolicy().hasHeightForWidth())
        self.in_name.setSizePolicy(sizePolicy)
        self.in_name.setMinimumSize(QSize(0, 0))
        self.in_name.setMaximumSize(QSize(16777215, 16777215))
        self.in_name.setAlignment(Qt.AlignCenter)

        self.verticalLayout.addWidget(self.in_name)

        self.c_axis = QComboBox(WAxisSelector)
        self.c_axis.setObjectName("c_axis")
        sizePolicy.setHeightForWidth(self.c_axis.sizePolicy().hasHeightForWidth())
        self.c_axis.setSizePolicy(sizePolicy)
        self.c_axis.setMinimumSize(QSize(0, 0))
        self.c_axis.setMaximumSize(QSize(16777215, 16777215))

        self.verticalLayout.addWidget(self.c_axis)

        self.c_action = QComboBox(WAxisSelector)
        self.c_action.addItem("")
        self.c_action.addItem("")
        self.c_action.addItem("")
        self.c_action.setObjectName("c_action")
        sizePolicy.setHeightForWidth(self.c_action.sizePolicy().hasHeightForWidth())
        self.c_action.setSizePolicy(sizePolicy)
        self.c_action.setMinimumSize(QSize(0, 0))
        self.c_action.setMaximumSize(QSize(16777215, 16777215))

        self.verticalLayout.addWidget(self.c_action)

        self.c_unit = QComboBox(WAxisSelector)
        self.c_unit.setObjectName("c_unit")
        sizePolicy.setHeightForWidth(self.c_unit.sizePolicy().hasHeightForWidth())
        self.c_unit.setSizePolicy(sizePolicy)
        self.c_unit.setMinimumSize(QSize(0, 0))
        self.c_unit.setMaximumSize(QSize(16777215, 16777215))

        self.verticalLayout.addWidget(self.c_unit)

        self.b_filter = QPushButton(WAxisSelector)
        self.b_filter.setObjectName("b_filter")
        sizePolicy.setHeightForWidth(self.b_filter.sizePolicy().hasHeightForWidth())
        self.b_filter.setSizePolicy(sizePolicy)
        self.b_filter.setMinimumSize(QSize(0, 0))
        self.b_filter.setMaximumSize(QSize(16777215, 16777215))

        self.verticalLayout.addWidget(self.b_filter)

        self.retranslateUi(WAxisSelector)

        QMetaObject.connectSlotsByName(WAxisSelector)

    # setupUi

    def retranslateUi(self, WAxisSelector):
        WAxisSelector.setWindowTitle("")
        self.in_name.setText(QCoreApplication.translate("WAxisSelector", "X", None))
        self.c_action.setItemText(
            0, QCoreApplication.translate("WAxisSelector", "None", None)
        )
        self.c_action.setItemText(
            1, QCoreApplication.translate("WAxisSelector", "FFT", None)
        )
        self.c_action.setItemText(
            2, QCoreApplication.translate("WAxisSelector", "Filter", None)
        )

        self.b_filter.setText(
            QCoreApplication.translate("WAxisSelector", "Filter", None)
        )

    # retranslateUi



from SciDataTool.Functions.Plot import ifft_dict, fft_dict


def is_axes_in_order(axes_selected, data):
    """Detect if axes given are in order compared to how they are store in a Data object
    Parameters
    ----------
    axes_selected: list
        List of axes selected as str
    data : Data
        Data object reference that store the axes
    Returns
    -------
    boolean stating if the axes are in order or not
    """
    not_in_order = False
    axes_name = [ax.name for ax in data.get_axes()]
    axes_selected_name = [ax.split("{")[0] for ax in axes_selected]

    if axes_selected_name[0] in axes_name and axes_selected_name[1] in axes_name:
        if axes_name.index(axes_selected_name[0]) > axes_name.index(
            axes_selected_name[1]
        ):
            not_in_order = True
            axes_selected = [axes_selected[1], axes_selected[0]]

    elif axes_selected_name[0] in ifft_dict and axes_selected_name[1] in ifft_dict:
        if axes_name.index(ifft_dict[axes_selected_name[0]]) > axes_name.index(
            ifft_dict[axes_selected_name[1]]
        ):

            not_in_order = True
            axes_selected = [axes_selected[1], axes_selected[0]]

    elif axes_selected_name[0] in fft_dict and axes_selected_name[1] in fft_dict:
        if axes_name.index(fft_dict[axes_selected_name[0]]) > axes_name.index(
            fft_dict[axes_selected_name[1]]
        ):

            not_in_order = True
            axes_selected = [axes_selected[1], axes_selected[0]]

    return not_in_order, axes_selected
